from ast import NameConstant
from typing import Any

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class BooleanPlugin(ParameterPlugin):

    COUNTABLE = True

    @classmethod
    def type(cls) -> Any:
        return bool

    @classmethod
    def type_name(cls) -> str:
        return "boolean"

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> bool:
        return super(BooleanPlugin, cls).parameter_value(type_defs, scene, project, action_id, parameter_id)

    @classmethod
    def _value_from_json(cls, value: str) -> bool:
        return super(BooleanPlugin, cls)._value_from_json(value)

    @classmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> NameConstant:
        return NameConstant(
            value=cls.parameter_execution_value(type_defs, scene, project, action_id, parameter_id), kind=None
        )


class BooleanListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return list[bool]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(BooleanPlugin)

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> list[bool]:
        return super(BooleanListPlugin, cls).parameter_value(type_defs, scene, project, action_id, parameter_id)



import asyncio
import time
from collections import deque
from typing import Any, Awaitable, Callable, Coroutine, TypeVar

import websockets
from aiologger.levels import LogLevel
from dataclasses_jsonschema import ValidationError
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import env, json
from arcor2.data.events import Event
from arcor2.data.rpc.common import RPC
from arcor2.exceptions import Arcor2Exception

MAX_RPC_DURATION = env.get_float("ARCOR2_MAX_RPC_DURATION", 0.1)

RPCT = TypeVar("RPCT", bound=RPC)
ReqT = TypeVar("ReqT", bound=RPC.Request)
RespT = TypeVar("RespT", bound=RPC.Response)

RPC_CB = Callable[[ReqT, WsClient], Coroutine[Any, Any, None | RespT]]
RPC_DICT_TYPE = dict[str, tuple[type[RPC], RPC_CB]]

EventT = TypeVar("EventT", bound=Event)
EVENT_DICT_TYPE = dict[str, tuple[type[EventT], Callable[[EventT, WsClient], Coroutine[Any, Any, None]]]]


def custom_exception_handler(loop: asyncio.AbstractEventLoop, context: dict[str, Any]) -> None:

    # it is also possible to use aiorun.run with stop_on_unhandled_errors=True but this prints much more useful info
    loop.default_exception_handler(context)

    if __debug__:  # stop loop while debugging, try to continue in production
        loop.stop()


async def send_json_to_client(client: WsClient, data: str) -> None:

    try:
        await client.send(data)
    except websockets.exceptions.ConnectionClosed:
        pass


async def server(
    client: Any,
    logger: Any,
    register: Callable[[Any], Awaitable[None]],
    unregister: Callable[[Any], Awaitable[None]],
    rpc_dict: RPC_DICT_TYPE,
    event_dict: None | EVENT_DICT_TYPE = None,
    verbose: bool = False,
) -> None:
    async def handle_message(msg: str) -> None:

        try:
            data = json.loads(msg)
        except json.JsonException as e:
            logger.error(f"Invalid data: '{msg}'.")
            logger.debug(e)
            return

        if not isinstance(data, dict):
            logger.error(f"Invalid data: '{data}'.")
            return

        if "request" in data:  # ...then it is RPC

            req_type = data["request"]

            try:
                rpc_cls, rpc_cb = rpc_dict[req_type]
            except KeyError:
                logger.error(f"Unknown RPC request: {data}.")
                return

            assert req_type == rpc_cls.__name__

            try:
                req = rpc_cls.Request.from_dict(data)
            except ValidationError as e:
                logger.error(f"Invalid RPC: {data}, error: {e}")
                return
            except Arcor2Exception as e:
                # this might happen if e.g. some dataclass does additional validation of values in its __post_init__
                try:
                    await client.send(rpc_cls.Response(data["id"], False, messages=[str(e)]).to_json())
                    logger.debug(e, exc_info=True)
                except (KeyError, websockets.exceptions.ConnectionClosed):
                    pass
                return

            else:

                try:
                    rpc_start = time.monotonic()
                    resp = await rpc_cb(req, client)
                    rpc_dur = time.monotonic() - rpc_start
                    if rpc_dur > MAX_RPC_DURATION:
                        logger.warn(f"{req.request} callback took {rpc_dur:.3f}s.")

                except Arcor2Exception as e:
                    logger.debug(e, exc_info=True)
                    resp = rpc_cls.Response(req.id, False, [str(e)])
                else:
                    if resp is None:  # default response
                        resp = rpc_cls.Response(req.id, True)
                    else:
                        assert isinstance(resp, rpc_cls.Response)
                        resp.id = req.id

            try:
                await client.send(resp.to_json())
            except websockets.exceptions.ConnectionClosed:
                return

            if logger.level == LogLevel.DEBUG:

                # Silencing of repetitive log messages
                # ...maybe this could be done better and in a more general way using logging.Filter?

                now = time.monotonic()
                if req.request not in req_last_ts:
                    req_last_ts[req.request] = deque()

                while req_last_ts[req.request]:
                    if req_last_ts[req.request][0] < now - 5.0:
                        req_last_ts[req.request].popleft()
                    else:
                        break

                req_last_ts[req.request].append(now)
                req_per_sec = len(req_last_ts[req.request]) / 5.0

                if req_per_sec > 2:
                    if req.request not in ignored_reqs:
                        ignored_reqs.add(req.request)
                        logger.debug(f"Request of type {req.request} will be silenced.")
                elif req_per_sec < 1:
                    if req.request in ignored_reqs:
                        ignored_reqs.remove(req.request)

                if req.request not in ignored_reqs:
                    logger.debug(f"RPC request: {req}, result: {resp}")

        elif "event" in data:  # ...event from UI

            assert event_dict

            try:
                event_cls, event_cb = event_dict[data["event"]]
            except KeyError as e:
                logger.error(f"Unknown event type: {e}.")
                return

            try:
                event = event_cls.from_dict(data)
            except ValidationError as e:
                logger.error(f"Invalid event: {data}, error: {e}")
                return

            await event_cb(event, client)

        else:
            logger.error(f"unsupported format of message: {data}")

    if event_dict is None:
        event_dict = {}

    req_last_ts: dict[str, deque] = {}
    ignored_reqs: set[str] = set()

    try:

        await register(client)

        loop = asyncio.get_event_loop()

        async for message in client:
            loop.create_task(handle_message(message))

    except websockets.exceptions.ConnectionClosed:
        pass
    finally:
        await unregister(client)



from arcor2.clients import asset
from arcor2.helpers import run_in_executor


async def asset_info() -> list[asset.AssetInfo]:
    return await run_in_executor(asset.asset_info)


async def asset_ids() -> set[str]:
    return await run_in_executor(asset.asset_ids)


async def delete_asset(id: str, remove_dependers: bool = False) -> None:
    await run_in_executor(asset.delete_asset, id, remove_dependers)


async def asset_exists(id: str) -> bool:
    return await run_in_executor(asset.asset_exists, id)



import time

from arcor2 import transformations as tr
from arcor2.data.common import Joint, Pose, StrEnum
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import MultiArmRobot, Settings


class DummyMultiArmRobot(MultiArmRobot):
    """Example class representing multiarm robot."""

    _ABSTRACT = False

    class Arms(StrEnum):
        left: str = "left"
        right: str = "right"

    GRIPPERS: dict[str, set[str]] = {
        Arms.left: {"l_gripper_1", "l_gripper_2"},
        Arms.right: {"r_gripper_1", "r_gripper_2"},
    }
    SUCTIONS: dict[str, set[str]] = {
        Arms.left: {"l_suction_1", "l_suction_2"},
        Arms.right: {"r_suction_1", "r_suction_2"},
    }
    EEF: dict[str, set[str]] = {Arms.left: {"l_eef_1", "l_eef_2"}, Arms.right: {"r_eef_1", "r_eef_2"}}

    JOINTS: dict[str, list[Joint]] = {Arms.left: [Joint("l_joint_1", 1.0)], Arms.right: [Joint("r_joint_1", -1.0)]}

    def __init__(self, obj_id: str, name: str, pose: Pose, settings: None | Settings = None) -> None:
        super().__init__(obj_id, name, pose, settings)

        self._hand_teaching: dict[str, bool] = {self.Arms.left: False, self.Arms.right: False}
        self._poses: dict[str, dict[str, Pose]] = {}

        for arm, eefs in self.EEF.items():
            self._poses[arm] = {}
            for eef in eefs:
                self._poses[arm][eef] = Pose()

    def get_arm_ids(self) -> set[str]:
        return self.Arms.set()

    @staticmethod
    def _get_from_dict(d: dict[str, set[str]], arm_id: None | str = None) -> set[str]:

        if arm_id is None:
            raise Arcor2Exception("Arm has to be specified.")

        try:
            return d[arm_id]
        except KeyError:
            raise Arcor2Exception("Unknown arm.")

    def get_end_effectors_ids(self, arm_id: None | str = None) -> set[str]:
        return self._get_from_dict(self.EEF, arm_id)

    def get_end_effector_pose(self, end_effector: str, arm_id: None | str = None) -> Pose:

        if end_effector not in self.get_end_effectors_ids(arm_id):
            raise Arcor2Exception("Unknown end effector.")

        assert arm_id

        return tr.make_pose_abs(self.pose, self._poses[arm_id][end_effector])

    def robot_joints(self, include_gripper: bool = False, arm_id: None | str = None) -> list[Joint]:
        """With no arm specified, returns all robot joints. Otherwise, returns
        joints for the given arm.

        :param arm_id:
        :return:
        """

        if arm_id is None:
            return [item for sublist in self.JOINTS.values() for item in sublist]

        try:
            return self.JOINTS[arm_id]
        except KeyError:
            raise Arcor2Exception("Unknown arm.")

    def grippers(self, arm_id: None | str = None) -> set[str]:
        return self._get_from_dict(self.GRIPPERS, arm_id)

    def suctions(self, arm_id: None | str = None) -> set[str]:
        return self._get_from_dict(self.SUCTIONS, arm_id)

    def move_to_pose(
        self,
        end_effector_id: str,
        target_pose: Pose,
        speed: float,
        safe: bool = True,
        linear: bool = True,
        arm_id: None | str = None,
    ) -> None:
        """Move given robot's end effector to the selected pose.

        :param end_effector_id:
        :param target_pose:
        :param speed:
        :param safe:
        :return:
        """

        if end_effector_id not in self.get_end_effectors_ids(arm_id):
            raise Arcor2Exception("Unknown end effector.")

        assert arm_id

        speed = min(max(0.0, speed), 1.0)

        with self._move_lock:
            time.sleep(1.0 - speed)
            self._poses[arm_id][end_effector_id] = tr.make_pose_rel(self.pose, target_pose)

    def move_to_joints(
        self, target_joints: list[Joint], speed: float, safe: bool = True, arm_id: None | str = None
    ) -> None:
        """Sets target joint values.

        :param target_joints:
        :param speed:
        :param safe:
        :return:
        """

        if arm_id is None:
            if len(target_joints) != len(self.robot_joints()):
                raise Arcor2Exception("Joints for both arms have to be specified.")
        elif arm_id not in self.get_arm_ids():
            raise Arcor2Exception("Unknown arm.")

        speed = min(max(0.0, speed), 1.0)

        with self._move_lock:
            time.sleep(1.0 - speed)

    def inverse_kinematics(
        self,
        end_effector_id: str,
        pose: Pose,
        start_joints: None | list[Joint] = None,
        avoid_collisions: bool = True,
        arm_id: None | str = None,
    ) -> list[Joint]:
        """Computes inverse kinematics.

        :param end_effector_id: IK target pose end-effector
        :param pose: IK target pose
        :param start_joints: IK start joints
        :param avoid_collisions: Return non-collision IK result if true
        :return: Inverse kinematics
        """

        if end_effector_id not in self.get_end_effectors_ids(arm_id):
            raise Arcor2Exception("Unknown end effector.")

        assert arm_id is not None

        try:
            return self.JOINTS[arm_id]
        except KeyError:
            raise Arcor2Exception("Unknown arm.")

    def forward_kinematics(self, end_effector_id: str, joints: list[Joint], arm_id: None | str = None) -> Pose:
        """Computes forward kinematics.

        :param end_effector_id: Target end effector name
        :param joints: Input joint values
        :return: Pose of the given end effector
        """

        if end_effector_id not in self.get_end_effectors_ids(arm_id):
            raise Arcor2Exception("Unknown end effector.")

        return Pose()

    def get_hand_teaching_mode(self, arm_id: None | str = None) -> bool:
        """
        This is expected to be implemented if the robot supports set_hand_teaching_mode
        :return:
        """

        if arm_id is None:
            raise Arcor2Exception("Arm has to be specified.")

        try:
            return self._hand_teaching[arm_id]
        except KeyError:
            raise Arcor2Exception("Unknown arm.")

    def set_hand_teaching_mode(self, enabled: bool, arm_id: None | str = None) -> None:

        if arm_id is None:
            raise Arcor2Exception("Arm has to be specified.")

        try:
            state = self._hand_teaching[arm_id]
        except KeyError:
            raise Arcor2Exception("Unknown arm.")

        if state == enabled:
            raise Arcor2Exception("That's the current state.")

        self._hand_teaching[arm_id] = enabled



import json
import logging
import traceback
from dataclasses import dataclass, field
from typing import Optional, Type, TypeAlias, Union, cast

from apispec import APISpec
from apispec_webframeworks.flask import FlaskPlugin
from dataclasses_jsonschema import DEFAULT_SCHEMA_TYPE, FieldMeta, JsonSchemaMixin
from dataclasses_jsonschema.apispec import DataclassesPlugin
from flask import Flask, Response, jsonify
from flask_cors import CORS
from flask_swagger_ui import get_swaggerui_blueprint

from arcor2 import __name__ as package_name
from arcor2 import env
from arcor2.data.common import WebApiError as IWebApiError
from arcor2.exceptions import Arcor2Exception

# TODO mypy likes Union here :-/
RespT: TypeAlias = Union[Response, tuple[str, int], tuple[Response, int]]  # noqa:NU001


debug = env.get_bool("ARCOR2_FLASK_DEBUG")


class FlaskException(Arcor2Exception):
    """Service should get WebApiError class and pass it to swagger
    generator."""

    service: str
    description: str

    def __init__(self, message: str, content: None | str = None):
        super().__init__(message)
        self.content = content

    @property
    def type(self):  # Use class.__name__ to get type when dealing with class objects
        return self.__class__.__name__

    def to_dict(self) -> dict:
        return IWebApiError(self.service, str(self), self.type, self.description, self.content).to_dict()


class General(FlaskException):
    service = package_name
    description = "Occurs when exception has no additional information or is unexpected."


class WebApiErrorFactory:
    """Build concrete web api schema class with correct fields description."""

    @staticmethod
    def _get_error_type_table(*allowed_exceptions: Type[FlaskException]) -> str:
        return "| Error type | Description |\n| --- | --- |\n" + "\n".join(
            [f"| {e.__name__} | {e.description} |" for e in allowed_exceptions]
        )

    @classmethod
    def get_class(cls, *allowed_exceptions: Type[FlaskException]) -> Type[IWebApiError]:
        # add general error to allowed for all services (this way it will appear in swagger)
        allowed_exceptions = (General, *allowed_exceptions)

        @dataclass
        class WebApiError(JsonSchemaMixin):
            """Represents any Web API error."""

            service: str = field(
                metadata=FieldMeta(
                    schema_type=DEFAULT_SCHEMA_TYPE,
                    description="Gets the service name where error occurred.",
                ).as_dict
            )

            message: str = field(
                metadata=FieldMeta(
                    schema_type=DEFAULT_SCHEMA_TYPE,
                    description="Gets human-readable explanation specific to this occurrence of the problem.",
                ).as_dict
            )

            type: str = field(
                metadata=FieldMeta(
                    schema_type=DEFAULT_SCHEMA_TYPE,
                    description=(
                        f"Gets application specific error type.\n\n" f"{cls._get_error_type_table(*allowed_exceptions)}"
                    ),
                ).as_dict
            )

            description: str = field(
                metadata=FieldMeta(
                    schema_type=DEFAULT_SCHEMA_TYPE,
                    description="Gets short, human-readable description of the application specific type.",
                ).as_dict
            )

            # This property should be parsed as JSON and validated against some model.
            # Validation and parsing should be handled by information consumer.
            # Model could be obtained from `type` property description.
            content: Optional[str] = field(
                metadata=FieldMeta(
                    schema_type=DEFAULT_SCHEMA_TYPE,
                    description=(
                        "Gets specific JSON object to this occurrence of the problem. "
                        "The data model is specific to the type "
                        "and should be documented on each API method."
                    ),
                    default=None,
                ).as_dict
            )

        return cast(Type[IWebApiError], WebApiError)


def create_app(import_name: str) -> Flask:
    app = Flask(import_name)
    CORS(app)
    return app


def run_app(
    app: Flask,
    name: str,
    version: str,
    port: int,
    dataclasses: None | list[type[JsonSchemaMixin]] = None,
    print_spec: bool = False,
    dependencies: None | dict[str, str] = None,
    *,
    api_version: None | str = None,
) -> None:
    if not api_version:
        api_version = version

    spec = APISpec(
        title=name,
        version=api_version,
        openapi_version="3.0.2",
        plugins=[FlaskPlugin(), DataclassesPlugin()],
        info={
            "description": "Dependencies: none"
            if dependencies is None
            else f'Dependencies: {", ".join([f"{k} {v}" for k, v in dependencies.items()])}'
        },
    )

    if dataclasses is not None:
        for dc in dataclasses:
            spec.components.schema(dc.__name__, schema=dc)

    with app.test_request_context():
        for rule in app.url_map.iter_rules():
            if rule.endpoint != "static":
                spec.path(view=app.view_functions[rule.endpoint])

    if print_spec:
        print(spec.to_yaml())
        return

    @app.route("/healthz/ready")
    def health():
        return Response(status=200)

    @app.route("/swagger/api/swagger.json", methods=["GET"])
    def get_swagger() -> RespT:
        return jsonify(spec.to_dict())

    @app.errorhandler(Arcor2Exception)
    def handle_bad_request_general(e: Arcor2Exception) -> tuple[str, int]:
        if debug:
            print("".join(traceback.format_exception(None, e, e.__traceback__)))
        return json.dumps(General(str(e)).to_dict()), 500

    @app.errorhandler(FlaskException)
    def handle_bad_request_intentional(e: FlaskException) -> tuple[str, int]:
        if debug:
            print("".join(traceback.format_exception(None, e, e.__traceback__)))
        return json.dumps(e.to_dict()), 500

    SWAGGER_URL = "/swagger"

    swaggerui_blueprint = get_swaggerui_blueprint(
        SWAGGER_URL, "./api/swagger.json"  # Swagger UI static files will be mapped to '{SWAGGER_URL}/dist/'
    )

    # Register blueprint at URL
    app.register_blueprint(swaggerui_blueprint, url_prefix=SWAGGER_URL)

    if not env.get_bool("ARCOR2_REST_API_DEBUG", False):
        # turn off logging each endpoint call by default
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)

    app.run(host="0.0.0.0", port=port)



from arcor2.data.common import ActionMetadata
from arcor2.object_types.abstract import Generic


class ParamToReturn(Generic):

    _ABSTRACT = False

    def bool_param(self, param: bool, *, an: None | str = None) -> bool:
        return param

    def int_param(self, param: int, *, an: None | str = None) -> int:
        return param

    def double_param(self, param: float, *, an: None | str = None) -> float:
        return param

    bool_param.__action__ = ActionMetadata()  # type: ignore
    int_param.__action__ = ActionMetadata()  # type: ignore
    double_param.__action__ = ActionMetadata()  # type: ignore



import ast
import inspect
from ast import Attribute, Load, Name
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Optional, get_type_hints

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2 import json
from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import IntEnum
from arcor2.data.object_type import ParameterMeta
from arcor2.parameter_plugins.base import ImportTuple, ParameterPlugin, ParameterPluginException, TypesDict

# TODO move IntEnum definition here?


@dataclass
class IntegerEnumExtra(JsonSchemaMixin):

    allowed_values: Optional[set[Any]] = None


class IntegerEnumPlugin(ParameterPlugin):

    EXACT_TYPE = False
    COUNTABLE = True

    @classmethod
    def type(cls) -> Any:
        return IntEnum

    @classmethod
    def type_name(cls) -> str:
        return "integer_enum"

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:
        super(IntegerEnumPlugin, cls).meta(param_meta, action_method, action_node)

        ttype = get_type_hints(action_method)[param_meta.name]
        if not issubclass(ttype, cls.type()):
            raise ParameterPluginException(f"Type {ttype.__name__} is not subclass of {cls.type().__name__}.")

        param_meta.extra = IntegerEnumExtra(ttype.set()).to_json()

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Enum:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        obj_id, action_type = action.parse_type()
        obj_type_name = scene.object(obj_id).type
        try:
            obj_type = type_defs[obj_type_name]
        except KeyError:
            raise ParameterPluginException(f"Unknown object type {obj_type_name}.")

        try:
            method = getattr(obj_type, action_type)
        except AttributeError:
            raise ParameterPluginException(f"Object type {obj_type_name} does not have method {action_type}.")

        try:
            ttype = get_type_hints(method)[param.name]
        except KeyError:
            raise ParameterPluginException(f"Method {obj_type}/{method.__name__} does not have parameter {param.name}.")

        if not issubclass(ttype, cls.type()):
            raise ParameterPluginException(f"Type {ttype.__name__} is not subclass of {cls.type().__name__}.")

        try:
            return ttype(json.loads(param.value))
        except (ValueError, json.JsonException):
            raise ParameterPluginException(f"Parameter {parameter_id} of action {action.name} has invalid value.")

    @classmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Attribute:

        val = cls.parameter_value(type_defs, scene, project, action_id, parameter_id)

        return Attribute(value=Name(id=val.__class__.__name__, ctx=Load()), attr=val.name, ctx=Load())

    @classmethod
    def need_to_be_imported(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> None | list[ImportTuple]:

        enum_cls = cls.parameter_value(type_defs, scene, project, action_id, parameter_id).__class__
        # TODO does this work as expected in all cases?
        module = inspect.getmodule(enum_cls)
        if not module:
            raise ParameterPluginException("Failed to get the module.")

        # TODO enums are typically defined in the same module as a object type but could be def. elsewhere (arcor2.data)
        return [ImportTuple(f"object_types.{module.__name__.split('.')[-1]}", enum_cls.__name__)]



import ast
import inspect
import os
import shutil
from dataclasses import is_dataclass
from typing import Any, Callable, Iterator, get_type_hints

import typing_inspect
from dataclasses_jsonschema import JsonSchemaMixin, ValidationError

import arcor2
from arcor2 import json
from arcor2.data.common import ActionMetadata, Parameter
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic, Settings
from arcor2.source.utils import find_class_def, parse


class ObjectTypeException(Arcor2Exception):
    pass


def get_containing_module_sources(type_def: type[object]) -> str:
    """Returns sources of the whole containing module.

    ...whereas inspect.getsource(type_def) returns just source of the class itself
    :param type_def:
    :return:
    """

    with open(inspect.getfile(type_def), "r") as source_file:
        return source_file.read()


def check_object_type(type_def: type[Generic]) -> None:
    """Checks whether the object type source is a valid one.

    :param type_def:
    :return:
    """

    if not issubclass(type_def, Generic):
        raise ObjectTypeException("Not a subclass of Generic.")

    # it might happen that the code is ok, but can't be parsed e.g. due to unsupported placement of comment
    # parse_def is not enough here - there might be something unparseable outside of the ObjectType class itself
    parse(get_containing_module_sources(type_def))

    # TODO some more (simple) checks here?


def built_in_types() -> Iterator[tuple[str, type[Generic]]]:
    """Yields class name and class definition tuple."""

    for cls_name, cls_type in inspect.getmembers(arcor2.object_types.abstract, predicate=inspect.isclass):

        try:
            if not issubclass(cls_type, Generic):
                continue
        except TypeError:
            # TODO detect type-hinted generics (e.g. dict[str, str]) instead?
            # CancelDict / issubclass() arg 1 must be a class
            continue

        yield cls_name, cls_type


def get_built_in_type(name: str) -> type[Generic]:

    for bname, cls in built_in_types():
        if name == bname:
            return cls

    raise KeyError


def built_in_types_names() -> set[str]:
    return {type_name for type_name, _ in built_in_types()}


class DataError(Arcor2Exception):
    pass


# TODO settings_to_params


def settings_from_params(
    type_def: type[Generic], settings: list[Parameter], overrides: None | list[Parameter] = None
) -> Settings:
    """Constructs instance of Settings from two arrays of parameters (scene
    settings and project overrides).

    :param type_def:
    :param settings:
    :param overrides:
    :return:
    """

    if overrides is None:
        overrides = []

    final: dict[str, Parameter] = {s.name: s for s in settings}
    for over in overrides:
        if over.name not in final:
            raise Arcor2Exception("Invalid override.")

        if over.type != final[over.name].type:
            raise Arcor2Exception("Type mismatch.")

        final[over.name] = over

    settings_def = get_settings_def(type_def)
    settings_data: dict[str, Any] = {}

    settings_def_type_hints = get_type_hints(settings_def.__init__)

    for s in final.values():

        try:
            setting_def = settings_def_type_hints[s.name]
        except KeyError as e:
            raise Arcor2Exception(f"Unknown property {s.name}.") from e

        try:
            if issubclass(setting_def, JsonSchemaMixin):
                settings_data[s.name] = json.loads(s.value)
            else:
                settings_data[s.name] = setting_def(json.loads(s.value))
        except (json.JsonException, ValidationError) as e:
            raise Arcor2Exception(f"Parameter {s.name} has invalid value.") from e

    try:
        settings_cls = settings_def.from_dict(settings_data)
    except (ValueError, ValidationError) as e:
        raise Arcor2Exception("Validation of settings failed.") from e

    return settings_cls


def get_settings_def(type_def: type[Generic]) -> type[Settings]:
    """Get settings definition from object type definition.

    :param type_def:
    :return:
    """

    sig = inspect.signature(type_def.__init__)
    try:
        param = sig.parameters["settings"]
    except KeyError:
        raise Arcor2Exception("Type has no settings.")

    args = typing_inspect.get_args(param.annotation)

    # should support all of: 'Optional[Settings]', 'None | Settings', 'Settings, None'
    if typing_inspect.is_optional_type(param.annotation) or (
        typing_inspect.is_union_type(param.annotation) and None in args
    ):

        for arg in args:
            if arg != type(None):  # noqa: E721
                settings_cls = arg
                break
        else:
            raise Arcor2Exception("Can't find annotation for object settings.")
    else:
        settings_cls = param.annotation

    try:
        if not issubclass(settings_cls, Settings):
            raise Arcor2Exception(f"Settings have invalid type ({settings_cls.__name__}).")
    except TypeError:
        raise Arcor2Exception("Settings have invalid annotation.")

    if not is_dataclass(settings_cls):
        raise Arcor2Exception("Settings misses @dataclass decorator.")

    return settings_cls


def base_from_source(source: str | ast.AST, cls_name: str) -> list[str]:
    """Returns list, where the first element is name of the base class and
    others are mixins.

    :param source:
    :param cls_name:
    :return:
    """

    if isinstance(source, str):
        cls_def = find_class_def(cls_name, parse(source))
    else:
        cls_def = find_class_def(cls_name, source)

    if not cls_def.bases:
        return []

    ret: list[str] = []

    for base in reversed(cls_def.bases):

        assert isinstance(base, ast.Name)
        ret.append(base.id)

    return ret


def iterate_over_actions(
    type_def: type[Generic],
) -> Iterator[tuple[str, Callable[[Any,], Any]]]:

    for method_name, method in inspect.getmembers(type_def, inspect.isroutine):

        try:
            if not isinstance(method.__action__, ActionMetadata):
                continue
        except AttributeError:
            continue

        yield method_name, method


def prepare_object_types_dir(path: str, module: str) -> None:
    """Creates a fresh directory, where ObjectTypes will be placed.

    :param path:
    :param module:
    :return:
    """

    full_path = os.path.join(path, module)

    if os.path.exists(full_path):
        shutil.rmtree(full_path)

    os.makedirs(full_path)

    with open(os.path.join(full_path, "__init__.py"), "w"):
        pass


__all__ = [
    ObjectTypeException.__name__,
    built_in_types.__name__,
    get_built_in_type.__name__,
    built_in_types_names.__name__,
    DataError.__name__,
]



from ast import stmt
from typing import Any

from PIL.Image import Image

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.exceptions import Arcor2NotImplemented
from arcor2.image import image_from_json, image_to_json
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict


class ImagePlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return Image

    @classmethod
    def type_name(cls) -> str:
        return "image"

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Image:
        json_str = super(ImagePlugin, cls).parameter_value(type_defs, scene, project, action_id, parameter_id)
        return cls._value_from_json(json_str)

    @classmethod
    def _value_from_json(cls, value: str) -> Image:
        return image_from_json(value)

    @classmethod
    def value_to_json(cls, value: Image) -> str:
        return image_to_json(value)

    @classmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> stmt:
        raise Arcor2NotImplemented("It does not make much sense to have image as literal.")



import fastuuid as uuid

from arcor2.data.rpc import common


def get_id() -> int:
    """Generates unique ID suitable for RPC/orjson."""

    return uuid.uuid4().int % 2**32


__all__ = [
    common.__name__,
]



from __future__ import annotations

import abc
import copy
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, unique
from json import JSONEncoder
from typing import Any, Iterator, NamedTuple, Optional, TypeVar, cast

import fastuuid as uuid
import numpy as np
import quaternion
from dataclasses_jsonschema import DEFAULT_SCHEMA_TYPE, JsonSchemaMixin
from dataclasses_jsonschema.type_defs import JsonDict, SchemaType

from arcor2 import json
from arcor2.exceptions import Arcor2Exception


def uid(prefix: str) -> str:

    if not (prefix and prefix[0].isalpha() and prefix[-1] != "_"):
        raise Arcor2Exception(f"{prefix} is a invalid uid prefix.")

    return f"{prefix}_{uuid.uuid4().hex}"


@unique
class StrEnum(str, Enum):
    @classmethod
    def set(cls) -> set[str]:
        return set(map(lambda c: c.value, cls))  # type: ignore


@unique
class IntEnum(int, Enum):
    @classmethod
    def set(cls) -> set[int]:
        return set(map(lambda c: c.value, cls))  # type: ignore


class FlowTypes(StrEnum):

    DEFAULT: str = "default"


class LinkToActionOutput(NamedTuple):
    action_id: str
    flow_name: FlowTypes
    output_index: int


def parse_link(val: str) -> LinkToActionOutput:

    try:
        action_id, flow_name, output_idx_str = val.split("/")
        return LinkToActionOutput(action_id, FlowTypes(flow_name), int(output_idx_str))
    except ValueError as e:
        raise Arcor2Exception("Invalid link value.") from e


class DataClassEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:

        if isinstance(o, JsonSchemaMixin):
            return o.to_dict()

        return super().default(o)


@dataclass
class IterableIndexable(JsonSchemaMixin):
    def __getitem__(self, item: int) -> float:

        attr = getattr(self, tuple(self.__dict__.keys())[item])
        assert isinstance(attr, (float, int))
        return attr

    def __iter__(self) -> Iterator[float]:

        # filtering items starting with _ is necessary to allow runtime monkey-patching
        # ...otherwise those patched attributes will appear in e.g. list(Position(1,2,3))
        yield from [v for k, v in self.__dict__.items() if not k.startswith("_")]


@dataclass
class Position(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def rotated(self, ori: Orientation, inverse: bool = False) -> Position:

        q = ori.as_quaternion()

        if inverse:
            q = q.inverse()

        rotated_vector = quaternion.rotate_vectors([q], [list(self)])[0][0]

        return Position(rotated_vector[0], rotated_vector[1], rotated_vector[2])

    def __eq__(self, other: object) -> bool:

        if not isinstance(other, Position):
            return False

        return np.allclose(list(self), list(other), rtol=1.0e-6)

    def __add__(self, other: object) -> Position:

        if not isinstance(other, Position):
            raise Arcor2Exception("Not a position.")

        return Position(self.x + other.x, self.y + other.y, self.z + other.z)

    def __iadd__(self, other: object) -> Position:

        if not isinstance(other, Position):
            raise ValueError

        self.x += other.x
        self.y += other.y
        self.z += other.z
        return self

    def __sub__(self, other) -> Position:

        if not isinstance(other, Position):
            raise Arcor2Exception("Not a position.")

        return Position(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: object) -> Position:

        if not isinstance(other, (float, int)):
            raise ValueError

        return Position(self.x * other, self.y * other, self.z * other)

    def __imul__(self, other: object) -> Position:

        if not isinstance(other, (float, int)):
            raise ValueError

        self.x *= other
        self.y *= other
        self.z *= other

        return self

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,
        schema_type: SchemaType = DEFAULT_SCHEMA_TYPE,
    ) -> JsonDict:

        # orjson does not like numpy.float64
        return {"x": float(self.x), "y": float(self.y), "z": float(self.z)}


@dataclass
class Orientation(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 1.0

    @classmethod
    def from_rotation_vector(cls, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> Orientation:
        return cls.from_quaternion(quaternion.from_rotation_vector([x, y, z]))

    @classmethod
    def from_quaternion(cls, q: quaternion.quaternion) -> Orientation:
        return Orientation(q.x, q.y, q.z, q.w)

    @staticmethod
    def _normalized(q: quaternion.quaternion) -> quaternion.quaternion:

        nq = q.normalized()

        if nq.isnan():
            raise Arcor2Exception("Invalid quaternion.")

        return nq

    def inverse(self) -> None:
        self.set_from_quaternion(self.as_quaternion().inverse())

    def inversed(self) -> Orientation:
        return self.from_quaternion(self.as_quaternion().inverse())

    def as_quaternion(self) -> quaternion.quaternion:
        return self._normalized(quaternion.quaternion(self.w, self.x, self.y, self.z))

    def set_from_quaternion(self, q: quaternion.quaternion) -> None:

        nq = self._normalized(q)

        self.x = nq.x
        self.y = nq.y
        self.z = nq.z
        self.w = nq.w

    def as_tr_matrix(self) -> np.ndarray:
        """Returns 4x4 transformation matrix.

        :return:
        """

        arr = np.empty((4, 4))
        arr[:3, :3] = quaternion.as_rotation_matrix(self.as_quaternion())
        arr[3, :] = [0, 0, 0, 1]
        return arr

    def __eq__(self, other: object) -> bool:

        if not isinstance(other, Orientation):
            return False

        return cast(bool, quaternion.isclose(self.as_quaternion(), other.as_quaternion(), rtol=1.0e-6)[0])

    def __mul__(self, other: object) -> Orientation:

        if not isinstance(other, Orientation):
            raise ValueError

        return self.from_quaternion(self.as_quaternion() * other.as_quaternion())

    def __imul__(self, other: object) -> Orientation:

        if not isinstance(other, Orientation):
            raise ValueError

        self.set_from_quaternion(other.as_quaternion() * self.as_quaternion())
        return self

    def __post_init__(self):

        nq = self.as_quaternion()  # in order to get normalized quaternion

        self.x = nq.x
        self.y = nq.y
        self.z = nq.z
        self.w = nq.w

    def to_dict(
        self,
        omit_none: bool = True,
        validate: bool = False,
        validate_enums: bool = True,
        schema_type: SchemaType = DEFAULT_SCHEMA_TYPE,
    ) -> JsonDict:

        # orjson does not like numpy.float64
        return {"x": float(self.x), "y": float(self.y), "z": float(self.z), "w": float(self.w)}


class ModelMixin(abc.ABC):
    """Mixin for objects with 'id' property that is uuid."""

    id: str

    @classmethod
    @abc.abstractmethod
    def uid_prefix(cls) -> str:
        pass

    @classmethod
    def uid(cls) -> str:
        """Returns uid with proper prefix.

        :return:
        """
        return uid(cls.uid_prefix())

    def __post_init__(self) -> None:
        if not self.id:
            self.id = self.uid()


@dataclass
class NamedOrientation(JsonSchemaMixin, ModelMixin):

    name: str
    orientation: Orientation
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "ori"

    def copy(self) -> NamedOrientation:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class Pose(JsonSchemaMixin):

    position: Position = field(default_factory=Position)
    orientation: Orientation = field(default_factory=Orientation)

    def as_tr_matrix(self) -> np.ndarray:

        arr = np.empty((4, 4))
        arr[:3, :3] = quaternion.as_rotation_matrix(self.orientation.as_quaternion())
        arr[:3, 3] = list(self.position)
        arr[3, :] = [0, 0, 0, 1]
        return arr

    @staticmethod
    def from_tr_matrix(matrix: np.ndarray) -> Pose:

        tvec = matrix[:3, 3]
        return Pose(
            Position(tvec[0], tvec[1], tvec[2]),
            Orientation.from_quaternion(quaternion.from_rotation_matrix(matrix[:3, :3])),
        )

    def inversed(self) -> Pose:

        inv = self.orientation.inversed()
        return Pose((self.position * -1).rotated(inv), inv)


@dataclass
class ActionMetadata(JsonSchemaMixin):

    composite: bool = field(metadata=dict(description="Should be set for nested actions."), default=False)
    hidden: bool = field(metadata=dict(description="When set, action will be hidden in UIs."), default=False)
    cancellable: bool = field(
        init=False, default=False, metadata=dict(description="Defines whether action execution can be cancelled.")
    )


@dataclass
class Joint(JsonSchemaMixin):

    name: str
    value: float


@dataclass
class ProjectRobotJoints(JsonSchemaMixin, ModelMixin):

    name: str
    robot_id: str
    joints: list[Joint]
    is_valid: bool = False
    arm_id: Optional[str] = None
    end_effector_id: Optional[str] = None
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "joi"

    def copy(self) -> ProjectRobotJoints:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class Parameter(JsonSchemaMixin):

    name: str
    type: str
    value: str


@dataclass
class SceneObject(JsonSchemaMixin, ModelMixin):

    name: str
    type: str
    pose: Optional[Pose] = None
    parameters: list[Parameter] = field(default_factory=list)
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "obj"

    def copy(self) -> SceneObject:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class BareScene(JsonSchemaMixin, ModelMixin):

    name: str
    description: str = field(default_factory=str)
    created: Optional[datetime] = None
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "scn"

    SCN = TypeVar("SCN", bound="BareScene")

    def copy(self: SCN) -> SCN:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class Scene(BareScene):

    objects: list[SceneObject] = field(default_factory=list)

    @staticmethod
    def from_bare(bare: BareScene) -> Scene:
        return Scene(bare.name, bare.description, bare.created, bare.modified, bare.int_modified, id=bare.id)


@dataclass
class IdValue(JsonSchemaMixin):

    id: str
    value: str


class ActionParameterException(Arcor2Exception):
    pass


@dataclass
class ActionParameter(Parameter):
    class TypeEnum(StrEnum):

        PROJECT_PARAMETER: str = "project_parameter"
        LINK: str = "link"

    def str_from_value(self) -> str:

        val = json.loads(self.value)

        if not isinstance(val, str):
            raise Arcor2Exception("Value should be string.")

        return val

    def parse_link(self) -> LinkToActionOutput:
        if self.type != self.TypeEnum.LINK:
            raise Arcor2Exception("Not a link.")

        return parse_link(self.str_from_value())

    def is_value(self) -> bool:
        return self.type not in self.TypeEnum.set()


@dataclass
class Flow(JsonSchemaMixin):

    type: FlowTypes = FlowTypes.DEFAULT
    outputs: list[str] = field(default_factory=list)  # can't be set as it is unordered

    def __post_init__(self) -> None:

        if len(self.outputs) > len(set(self.outputs)):
            raise Arcor2Exception("Outputs have to be unique.")


@dataclass
class BareAction(JsonSchemaMixin, ModelMixin):

    name: str
    type: str
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "act"

    ACT = TypeVar("ACT", bound="BareAction")

    def copy(self: ACT) -> ACT:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class Action(BareAction):
    class ParsedType(NamedTuple):

        obj_id: str
        action_type: str

    parameters: list[ActionParameter] = field(default_factory=list)
    flows: list[Flow] = field(default_factory=list)

    def parse_type(self) -> ParsedType:

        try:
            obj_id_str, action = self.type.split("/")
        except ValueError:
            raise Arcor2Exception(f"Action: {self.id} has invalid type: {self.type}.")
        return self.ParsedType(obj_id_str, action)

    def parameter(self, parameter_id: str) -> ActionParameter:

        for param in self.parameters:
            if parameter_id == param.name:
                return param

        raise Arcor2Exception("Param not found")

    @property
    def bare(self) -> BareAction:
        return BareAction(self.name, self.type, id=self.id)

    def flow(self, flow_type: FlowTypes = FlowTypes.DEFAULT) -> Flow:

        for flow in self.flows:
            if flow.type == flow_type:
                return flow
        raise Arcor2Exception(f"Flow '{flow_type.value}' not found.")


@dataclass
class BareActionPoint(JsonSchemaMixin, ModelMixin):

    name: str
    position: Position
    parent: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "acp"

    ACP = TypeVar("ACP", bound="BareActionPoint")

    def copy(self: ACP) -> ACP:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class ActionPoint(BareActionPoint):

    orientations: list[NamedOrientation] = field(default_factory=list)
    robot_joints: list[ProjectRobotJoints] = field(default_factory=list)
    actions: list[Action] = field(default_factory=list)

    @staticmethod
    def from_bare(bare: BareActionPoint) -> ActionPoint:
        return ActionPoint(bare.name, bare.position, bare.parent, bare.display_name, bare.description, id=bare.id)


@dataclass
class ProjectLogicIf(JsonSchemaMixin):

    what: str
    value: str

    def parse_what(self) -> LinkToActionOutput:
        return parse_link(self.what)


@dataclass
class LogicItem(JsonSchemaMixin, ModelMixin):
    class ParsedStart(NamedTuple):

        start_action_id: str
        start_flow: str

    # TODO can't be ClassVar because of https://github.com/s-knibbs/dataclasses-jsonschema/issues/176
    START = "START"
    END = "END"

    start: str
    end: str
    condition: Optional[ProjectLogicIf] = None
    id: str = ""

    def parse_start(self) -> ParsedStart:

        try:
            start_action_id, start_flow = self.start.split("/")
        except ValueError:
            return self.ParsedStart(self.start, FlowTypes.DEFAULT)

        return self.ParsedStart(start_action_id, start_flow)

    @classmethod
    def uid_prefix(cls) -> str:
        return "lit"

    def copy(self) -> LogicItem:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class Range(JsonSchemaMixin):
    # TODO when nested in ProjectParameter, ProjectParameter.from_dict({}) raises NameError: name 'Range' is not defined

    min: float
    max: float


@dataclass
class ProjectParameter(Parameter, ModelMixin):

    range: Optional[Range] = None  # TODO use it in parameter plugins?
    display_name: Optional[str] = None
    description: Optional[str] = None
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "pco"


@dataclass
class FunctionReturns(JsonSchemaMixin):

    type: str
    link: str


@dataclass
class ProjectFunction(JsonSchemaMixin, ModelMixin):

    name: str
    actions: list[Action] = field(default_factory=list)
    logic: list[LogicItem] = field(default_factory=list)
    parameters: list[ActionParameter] = field(default_factory=list)
    returns: list[FunctionReturns] = field(default_factory=list)
    id: str = ""

    def action_ids(self) -> set[str]:
        return {act.id for act in self.actions}

    def action(self, action_id: str) -> Action:

        for ac in self.actions:
            if ac.id == action_id:
                return ac
        else:
            raise Arcor2Exception("Action not found")

    @classmethod
    def uid_prefix(cls) -> str:
        return "pfu"

    def copy(self) -> ProjectFunction:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class SceneObjectOverride(JsonSchemaMixin):

    id: str  # object id
    parameters: list[Parameter]


@dataclass
class BareProject(JsonSchemaMixin, ModelMixin):

    name: str
    scene_id: str
    description: str = field(default_factory=str)
    has_logic: bool = True
    created: Optional[datetime] = None
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None
    id: str = ""

    @classmethod
    def uid_prefix(cls) -> str:
        return "pro"

    PRO = TypeVar("PRO", bound="BareProject")

    def copy(self: PRO) -> PRO:
        c = copy.deepcopy(self)
        c.id = self.uid()
        return c


@dataclass
class Project(BareProject):

    action_points: list[ActionPoint] = field(default_factory=list)
    parameters: list[ProjectParameter] = field(default_factory=list)
    functions: list[ProjectFunction] = field(default_factory=list)
    logic: list[LogicItem] = field(default_factory=list)
    object_overrides: list[SceneObjectOverride] = field(default_factory=list)
    project_objects_ids: Optional[list[str]] = None  # not used at the moment

    @staticmethod
    def from_bare(bare: BareProject) -> Project:
        return Project(
            bare.name,
            bare.scene_id,
            bare.description,
            bare.has_logic,
            bare.created,
            bare.modified,
            bare.int_modified,
            id=bare.id,
        )


@dataclass
class ProjectSources(JsonSchemaMixin):

    id: str  # project_id
    script: str


@dataclass
class IdDesc(JsonSchemaMixin):
    id: str
    name: str
    created: datetime
    modified: datetime
    description: Optional[str] = None


@dataclass
class BroadcastInfo(JsonSchemaMixin):

    host: str
    port: int


@dataclass
class Error(JsonSchemaMixin):
    code: int
    detail: str
    type: Optional[str] = None
    data: Optional[str] = None


@dataclass
class WebApiError(JsonSchemaMixin, Arcor2Exception):
    service: str
    message: str
    type: str
    description: str
    content: Optional[str] = None

    def __str__(self):
        return f"{self.service} ({self.type}): {self.message}"



from __future__ import annotations

import copy
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Iterator, ValuesView

from arcor2.data import common as cmn
from arcor2.exceptions import Arcor2Exception

if __debug__:
    import random

# TODO cached ProjectFunction (actions from functions are totally ignored at the moment)


class CachedSceneException(Arcor2Exception):
    pass


class CachedBase:

    __slots__ = "id", "name", "description", "created", "modified", "int_modified"

    def __init__(self, data: cmn.Scene | cmn.Project | CachedScene | CachedProject) -> None:

        self.id: str = data.id
        self.name: str = data.name
        self.description: str = data.description

        self.created: None | datetime = data.created
        self.modified: None | datetime = data.modified
        self.int_modified: None | datetime = data.int_modified


class UpdateableMixin:

    if TYPE_CHECKING:
        __slots__ = "modified", "int_modified"

        modified: None | datetime = None
        int_modified: None | datetime = None
    else:
        __slots__ = ()

    def update_modified(self) -> None:
        self.int_modified = datetime.now(tz=timezone.utc)

    @property
    def has_changes(self) -> bool:
        """Returns whether the scene/project has some unsaved changes.

        :return:
        """

        # this is true for a newly created scene/project
        # it does not matter if there are changes or not - the scene/project was not saved yet
        if self.modified is None:
            return True

        # this is true for scene/project that was loaded but there are not changes yet
        if self.int_modified is None:
            return False

        # if the scene/project was already saved once (modified is not None)
        # and there were some changes (int_modified is not None)
        # let's compare if some change happened (int_modified) after saving to the Project service (modified)
        return self.int_modified > self.modified


class CachedScene(CachedBase):

    __slots__ = ("_objects", "_object_types")

    def __init__(self, scene: cmn.Scene | CachedScene) -> None:

        super().__init__(scene)

        self._objects: dict[str, cmn.SceneObject] = {}

        if isinstance(scene, CachedScene):
            self._objects = scene._objects
        else:
            # TODO deal with children

            for obj in scene.objects:

                if obj.id in self._objects:
                    raise CachedSceneException(f"Duplicate object id: {obj.id}.")

                self._objects[obj.id] = obj

        self._object_types = {obj.type for obj in self.objects}

    @property
    def bare(self) -> cmn.BareScene:
        return cmn.BareScene(self.name, self.description, self.created, self.modified, self.int_modified, id=self.id)

    def object_names(self) -> Iterator[str]:

        for obj in self._objects.values():
            yield obj.name

    @property
    def objects(self) -> Iterator[cmn.SceneObject]:

        for obj in self._objects.values():
            yield obj

    @property
    def object_ids(self) -> set[str]:
        return set(self._objects)

    def object(self, object_id: str) -> cmn.SceneObject:

        try:
            return self._objects[object_id]
        except KeyError:
            raise Arcor2Exception(f"Object ID {object_id} not found.")

    def objects_of_type(self, obj_type: str) -> Iterator[cmn.SceneObject]:

        for obj in self.objects:
            if obj.type == obj_type:
                yield obj

    @property
    def object_types(self) -> set[str]:
        assert self._object_types == {obj.type for obj in self.objects}
        return self._object_types

    @property
    def scene(self) -> cmn.Scene:

        sc = cmn.Scene.from_bare(self.bare)
        sc.modified = self.modified
        sc.int_modified = self.int_modified
        sc.objects = list(self.objects)
        return sc


class UpdateableCachedScene(UpdateableMixin, CachedScene):

    __slots__ = ()

    def __init__(self, scene: cmn.Scene | CachedScene):
        super(UpdateableCachedScene, self).__init__(copy.deepcopy(scene))

    def upsert_object(self, obj: cmn.SceneObject) -> None:

        self._objects[obj.id] = obj
        self._object_types.add(obj.type)
        self.update_modified()

    def delete_object(self, obj_id: str) -> None:

        try:
            obj = self._objects.pop(obj_id)
        except KeyError as e:
            raise Arcor2Exception("Object id not found.") from e

        if not any(True for _ in self.objects_of_type(obj.type)):
            self._object_types.discard(obj.type)

        assert self._object_types == {obj.type for obj in self.objects}

        self.update_modified()


class CachedProjectException(Arcor2Exception):
    pass


@dataclass
class Parent:

    __slots__ = ("ap",)

    ap: cmn.BareActionPoint


@dataclass
class ApAction(Parent):

    __slots__ = ("action",)

    action: cmn.Action


@dataclass
class ApJoints(Parent):

    __slots__ = ("joints",)

    joints: cmn.ProjectRobotJoints


@dataclass
class ApOrientation(Parent):

    __slots__ = ("orientation",)

    orientation: cmn.NamedOrientation


class CachedProject(CachedBase):

    __slots__ = (
        "scene_id",
        "has_logic",
        "_action_points",
        "_actions",
        "_joints",
        "_orientations",
        "_parameters",
        "_logic_items",
        "_functions",
        "overrides",
        "_childs",
        "project_objects_ids",
    )

    def __init__(self, project: cmn.Project | CachedProject):

        super().__init__(project)

        self.scene_id: str = project.scene_id
        self.has_logic: bool = project.has_logic
        self.project_objects_ids: None | list[str] = project.project_objects_ids

        self._action_points: dict[str, cmn.BareActionPoint] = {}

        self._actions: dict[str, ApAction] = {}
        self._joints: dict[str, ApJoints] = {}
        self._orientations: dict[str, ApOrientation] = {}

        self._parameters: dict[str, cmn.ProjectParameter] = {}
        self._logic_items: dict[str, cmn.LogicItem] = {}
        self._functions: dict[str, cmn.ProjectFunction] = {}

        self.overrides: dict[str, list[cmn.Parameter]] = {}

        self._childs: dict[str, set[str]] = {}

        if isinstance(project, CachedProject):
            self._action_points = project._action_points
            self._actions = project._actions
            self._joints = project._joints
            self._orientations = project._orientations
            self._parameters = project._parameters
            self._logic_items = project._logic_items
            self._functions = project._functions
            self.overrides = project.overrides
            self._childs = project._childs

        else:

            for ap in project.action_points:

                if ap.id in self._action_points:
                    raise CachedProjectException(f"Duplicate AP id: {ap.id}.")

                bare_ap = cmn.BareActionPoint(ap.name, ap.position, ap.parent, id=ap.id)
                self._action_points[ap.id] = bare_ap
                self._upsert_child(ap.parent, ap.id)

                for ac in ap.actions:

                    if ac.id in self._actions:
                        raise CachedProjectException(f"Duplicate action id: {ac.id}.")

                    self._actions[ac.id] = ApAction(bare_ap, ac)
                    self._upsert_child(ap.id, ac.id)

                for joints in ap.robot_joints:

                    if joints.id in self._joints:
                        raise CachedProjectException(f"Duplicate joints id: {joints.id}.")

                    self._joints[joints.id] = ApJoints(bare_ap, joints)
                    self._upsert_child(ap.id, joints.id)

                for orientation in ap.orientations:

                    if orientation.id in self._orientations:
                        raise CachedProjectException(f"Duplicate orientation id: {orientation.id}.")

                    self._orientations[orientation.id] = ApOrientation(bare_ap, orientation)
                    self._upsert_child(ap.id, orientation.id)

            for override in project.object_overrides:
                self.overrides[override.id] = override.parameters

            for param in project.parameters:
                self._parameters[param.id] = param

            for logic_item in project.logic:
                self._logic_items[logic_item.id] = logic_item

            for function in project.functions:
                self._functions[function.id] = function

    @property
    def logic(self) -> ValuesView[cmn.LogicItem]:
        return self._logic_items.values()

    @property
    def valid_logic_endpoints(self) -> set[str]:
        return {cmn.LogicItem.START, cmn.LogicItem.END} | self._logic_items.keys()

    @property
    def parameters(self) -> ValuesView[cmn.ProjectParameter]:
        return self._parameters.values()

    @property
    def parameters_ids(self) -> set[str]:
        return set(self._parameters)

    @property
    def functions(self) -> ValuesView[cmn.ProjectFunction]:
        return self._functions.values()

    @property
    def project(self) -> cmn.Project:

        proj = cmn.Project.from_bare(self.bare)

        for bare_ap in self._action_points.values():

            ap = cmn.ActionPoint.from_bare(bare_ap)
            ap.actions = self.ap_actions(ap.id)
            ap.robot_joints = self.ap_joints(ap.id)
            ap.orientations = self.ap_orientations(ap.id)
            proj.action_points.append(ap)

        proj.object_overrides = [cmn.SceneObjectOverride(k, v) for k, v in self.overrides.items()]
        proj.parameters = list(self.parameters)
        proj.functions = list(self.functions)
        proj.logic = list(self.logic)
        proj.project_objects_ids = self.project_objects_ids
        return proj

    @property
    def bare(self) -> cmn.BareProject:
        return cmn.BareProject(
            self.name,
            self.scene_id,
            self.description,
            self.has_logic,
            self.created,
            self.modified,
            self.int_modified,
            id=self.id,
        )

    @property
    def action_points(self) -> ValuesView[cmn.BareActionPoint]:
        return self._action_points.values()

    @property
    def action_points_with_parent(self) -> list[cmn.BareActionPoint]:
        """Get action points which are relative to something (parent is set).

        :return:
        """

        return [ap for ap in self._action_points.values() if ap.parent]

    @property
    def action_names(self) -> set[str]:
        return {value.action.name for value in self._actions.values()}

    @property
    def action_points_names(self) -> set[str]:
        return {ap.name for ap in self._action_points.values()}

    @property
    def action_points_ids(self) -> set[str]:
        return set(self._action_points)

    def ap_and_joints(self, joints_id: str) -> tuple[cmn.BareActionPoint, cmn.ProjectRobotJoints]:

        try:
            value = self._joints[joints_id]
        except KeyError:
            raise CachedProjectException("Unknown joints.")

        return value.ap, value.joints

    def joints(self, joints_id: str) -> cmn.ProjectRobotJoints:

        try:
            return self._joints[joints_id].joints
        except KeyError:
            raise CachedProjectException("Unknown joints.")

    def bare_ap_and_orientation(self, orientation_id: str) -> tuple[cmn.BareActionPoint, cmn.NamedOrientation]:

        try:
            value = self._orientations[orientation_id]
        except KeyError:
            raise CachedProjectException("Unknown orientation.")

        return value.ap, value.orientation

    def pose(self, orientation_id: str) -> cmn.Pose:

        ap, ori = self.bare_ap_and_orientation(orientation_id)
        return cmn.Pose(ap.position, ori.orientation)

    def ap_orientations(self, ap_id: str) -> list[cmn.NamedOrientation]:

        return [value.orientation for value in self._orientations.values() if ap_id == value.ap.id]

    def ap_joints(self, ap_id: str) -> list[cmn.ProjectRobotJoints]:

        return [value.joints for value in self._joints.values() if ap_id == value.ap.id]

    def ap_actions(self, ap_id: str) -> list[cmn.Action]:

        return [value.action for value in self._actions.values() if ap_id == value.ap.id]

    def ap_action_ids(self, ap_id: str) -> set[str]:
        return {ac.id for ac in self.ap_actions(ap_id)}

    def ap_orientation_names(self, ap_id: str) -> set[str]:
        return {ori.name for ori in self.ap_orientations(ap_id)}

    def ap_joint_names(self, ap_id: str) -> set[str]:
        return {joints.name for joints in self.ap_joints(ap_id)}

    def ap_action_names(self, ap_id: str) -> set[str]:
        return {action.name for action in self.ap_actions(ap_id)}

    def orientation(self, orientation_id: str) -> cmn.NamedOrientation:

        try:
            return self._orientations[orientation_id].orientation
        except KeyError:
            raise CachedProjectException("Unknown orientation.")

    def action(self, action_id: str) -> cmn.Action:

        try:
            return self._actions[action_id].action
        except KeyError:
            raise CachedProjectException("Action not found")

    def action_io(self, action_id: str) -> tuple[list[cmn.LogicItem], list[cmn.LogicItem]]:
        """Returns list of logical connection ending in the action (its inputs)
        and starting from the action (its outputs)

        :param action_id:
        :return:
        """

        inputs: list[cmn.LogicItem] = []
        outputs: list[cmn.LogicItem] = []

        for item in self._logic_items.values():
            parsed_start = item.parse_start()
            if parsed_start.start_action_id == action_id:
                outputs.append(item)
            if item.end == action_id:
                inputs.append(item)

        if __debug__:  # make it a bit harder for tests to succeed
            random.shuffle(inputs)
            random.shuffle(outputs)

        return inputs, outputs

    def first_action_id(self) -> str:

        first_action: None | str = None

        for item in self._logic_items.values():
            if item.start == item.START:
                if first_action:
                    raise CachedProjectException("Duplicate start.")
                first_action = self.action(item.end).id

        if first_action is None:
            raise CachedProjectException("Start action not found.")

        return first_action

    def action_point_and_action(self, action_id: str) -> tuple[cmn.BareActionPoint, cmn.Action]:

        try:
            value = self._actions[action_id]
        except KeyError:
            raise CachedProjectException("Action not found")

        return value.ap, value.action

    @property
    def actions(self) -> list[cmn.Action]:
        return [value.action for value in self._actions.values()]

    def action_ids(self) -> set[str]:
        return set(self._actions)

    def bare_action_point(self, action_point_id: str) -> cmn.BareActionPoint:

        try:
            return self._action_points[action_point_id]
        except KeyError:
            raise CachedProjectException("Action point not found")

    def action_point(self, action_point_id: str) -> cmn.ActionPoint:

        ap = cmn.ActionPoint.from_bare(self.bare_action_point(action_point_id))
        ap.orientations = self.ap_orientations(action_point_id)
        ap.robot_joints = self.ap_joints(action_point_id)
        return ap

    def logic_item(self, logic_item_id: str) -> cmn.LogicItem:

        try:
            return self._logic_items[logic_item_id]
        except KeyError:
            raise CachedProjectException("LogicItem not found.")

    def parameter(self, parameter_id: str) -> cmn.ProjectParameter:
        """Gets a project parameter by its ID.

        :param parameter_id:
        :return:
        """

        try:
            return self._parameters[parameter_id]
        except KeyError:
            raise CachedProjectException("Project parameter not found.")

    def get_by_id(
        self, obj_id: str
    ) -> cmn.BareActionPoint | cmn.NamedOrientation | cmn.ProjectRobotJoints | cmn.Action | cmn.ProjectParameter:

        if obj_id in self._action_points:  # AP
            return self._action_points[obj_id]
        elif obj_id in self._joints:  # Joints
            return self._joints[obj_id].joints
        elif obj_id in self._orientations:  # Orientation
            return self._orientations[obj_id].orientation
        elif obj_id in self._actions:  # Action
            return self._actions[obj_id].action
        elif obj_id in self._parameters:
            return self._parameters[obj_id]

        raise CachedProjectException("Object not found.")

    def get_parent_id(self, obj_id: str) -> None | str:

        if obj_id in self._action_points:  # AP
            return self._action_points[obj_id].parent if self._action_points[obj_id].parent else None
        elif obj_id in self._joints:  # Joints
            return self._joints[obj_id].ap.id
        elif obj_id in self._orientations:  # Orientation
            return self._orientations[obj_id].ap.id
        elif obj_id in self._actions:  # Action
            return self._actions[obj_id].ap.id

        raise CachedProjectException("Object not found.")

    def _upsert_child(self, parent: None | str, child: str) -> None:
        if parent:
            if parent not in self._childs:
                self._childs[parent] = set()
            self._childs[parent].add(child)

    def childs(self, obj_id: str, recursive: bool = False) -> set[str]:

        try:
            ret = self._childs[obj_id]
        except KeyError:
            return set()  # TODO distinguish between no childs and unknown object

        if not recursive:
            return ret

        return ret | {c for s in [self.childs(ch, True) for ch in ret] for c in s}

    def _remove_child(self, parent: None | str, child: str) -> None:

        if parent and parent in self._childs:
            self._childs[parent].remove(child)
            if not self._childs[parent]:
                del self._childs[parent]

    def update_child(self, obj_id: str, old_parent: None | str, new_parent: None | str) -> None:

        self._remove_child(old_parent, obj_id)
        self._upsert_child(new_parent, obj_id)


class UpdateableCachedProject(UpdateableMixin, CachedProject):

    __slots__ = ()

    def __init__(self, project: cmn.Project | CachedProject):
        super().__init__(copy.deepcopy(project))

    def upsert_action(self, ap_id: str, action: cmn.Action) -> None:

        ap = self.bare_action_point(ap_id)

        if action.id in self._actions:
            assert self._actions[action.id].ap == ap
            self._actions[action.id].action = action
        else:
            self._actions[action.id] = ApAction(ap, action)

        self._upsert_child(ap.id, action.id)
        self.update_modified()

    def remove_action(self, action_id: str) -> cmn.Action:

        try:
            value = self._actions.pop(action_id)
        except KeyError as e:
            raise CachedProjectException("Action not found.") from e

        self._remove_child(value.ap.id, action_id)
        self.update_modified()
        return value.action

    def invalidate_joints(self, ap_id: str) -> None:

        for joints in self.ap_joints(ap_id):
            joints.is_valid = False

    def update_ap_position(self, ap_id: str, position: cmn.Position) -> None:

        ap = self.bare_action_point(ap_id)
        ap.position = position
        self.invalidate_joints(ap_id)
        self.update_modified()

    def upsert_orientation(self, ap_id: str, orientation: cmn.NamedOrientation) -> None:

        ap = self.bare_action_point(ap_id)

        if orientation.id in self._orientations:
            assert self._orientations[orientation.id].ap == ap
            self._orientations[orientation.id].orientation = orientation
        else:
            self._orientations[orientation.id] = ApOrientation(ap, orientation)

        self._upsert_child(ap_id, orientation.id)
        self.update_modified()

    def remove_orientation(self, orientation_id: str) -> cmn.NamedOrientation:

        try:
            value = self._orientations.pop(orientation_id)
        except KeyError as e:
            raise CachedProjectException("Orientation not found.") from e

        self._remove_child(value.ap.id, orientation_id)
        self.update_modified()
        return value.orientation

    def upsert_joints(self, ap_id: str, joints: cmn.ProjectRobotJoints) -> None:

        ap = self.bare_action_point(ap_id)

        if joints.id in self._joints:
            assert self._joints[joints.id].ap == ap
            self._joints[joints.id].joints = joints
        else:
            self._joints[joints.id] = ApJoints(ap, joints)

        self._upsert_child(ap_id, joints.id)
        self.update_modified()

    def remove_joints(self, joints_id: str) -> cmn.ProjectRobotJoints:

        try:
            value = self._joints.pop(joints_id)
        except KeyError as e:
            raise CachedProjectException("Joints not found.") from e

        self._remove_child(value.ap.id, joints_id)
        self.update_modified()
        return value.joints

    def upsert_action_point(
        self, ap_id: str, name: str, position: cmn.Position, parent: None | str = None
    ) -> cmn.BareActionPoint:

        try:
            ap = self.bare_action_point(ap_id)
            ap.name = name
            if position != ap.position:
                self.invalidate_joints(ap_id)
            ap.position = position
            ap.parent = parent
        except CachedProjectException:
            ap = cmn.BareActionPoint(name, position, parent, id=ap_id)
            self._action_points[ap_id] = ap
        self._upsert_child(parent, ap_id)
        self.update_modified()
        return ap

    def remove_action_point(self, ap_id: str) -> cmn.BareActionPoint:

        ap = self.bare_action_point(ap_id)

        for action in self.ap_actions(ap_id):
            self.remove_action(action.id)

        for joints in self.ap_joints(ap_id):
            self.remove_joints(joints.id)

        for ori in self.ap_orientations(ap_id):
            self.remove_orientation(ori.id)

        self._remove_child(ap.parent, ap_id)
        del self._action_points[ap_id]
        self.update_modified()
        return ap

    def upsert_logic_item(self, logic_item: cmn.LogicItem) -> None:

        self._logic_items[logic_item.id] = logic_item
        self.update_modified()

    def remove_logic_item(self, logic_item_id: str) -> cmn.LogicItem:

        try:
            logic_item = self._logic_items.pop(logic_item_id)
        except KeyError as e:
            raise CachedProjectException("Logic item not found.") from e
        self.update_modified()
        return logic_item

    def clear_logic(self) -> None:

        self._logic_items.clear()
        self.update_modified()

    def upsert_parameter(self, parameter: cmn.ProjectParameter) -> None:
        self._parameters[parameter.id] = parameter
        self.update_modified()

    def remove_parameter(self, parameter_id: str) -> cmn.ProjectParameter:

        try:
            param = self._parameters.pop(parameter_id)
        except KeyError as e:
            raise CachedProjectException("Project parameter not found.") from e
        self.update_modified()
        return param



import collections
import fractions
import io
import math
import os
import tempfile
import zipfile

from arcor2 import rest
from arcor2.exceptions import Arcor2Exception

# this is a workaround to make urdfpy/networkx happy under Python 3.9/3.10
# TODO not sure if it really works in runtime, the best would be to solve https://github.com/mmatl/urdfpy/issues/4
# or to switch to another package as proposed here https://github.com/robofit/arcor2/issues/706
fractions.gcd = math.gcd  # type: ignore

# another workaround related to urdfpy/networkx and Python 3.10
from collections.abc import Iterable, Mapping, Set  # noqa:E402

collections.Mapping = Mapping  # type: ignore
collections.Set = Set  # type: ignore
collections.Iterable = Iterable  # type: ignore

from urdfpy import URDF  # noqa


class Arcor2UrdfException(Arcor2Exception):
    pass


def is_urdf_file(fname: str) -> bool:

    _, ext = os.path.splitext(fname)
    return ext in (".urdf", ".xml")


def urdf_from_path(path_to_urdf: str) -> URDF:

    for file in os.listdir(path_to_urdf):
        if is_urdf_file(file):
            with open(os.path.join(path_to_urdf, file)) as f:
                s = f.read()
            break
    else:
        raise Arcor2UrdfException("Failed to find urdf file.")

    s = s.replace("package://", "")

    buff = io.BytesIO(s.encode())
    buff.name = os.path.join(path_to_urdf, file)

    try:
        return URDF.load(buff)
    except (ValueError, TypeError) as e:
        raise Arcor2UrdfException(str(e)) from e


def urdf_from_url(url_of_zipped_package: str) -> URDF:

    with tempfile.TemporaryDirectory() as tmp_dir:
        path_to_zip = os.path.join(tmp_dir, "urdf.zip")
        rest.download(url_of_zipped_package, path_to_zip)

        with zipfile.ZipFile(path_to_zip, "r") as zip_ref:
            zip_ref.extractall(tmp_dir)

        return urdf_from_path(tmp_dir)



from dataclasses import dataclass
from typing import Optional

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import Position
from arcor2.exceptions import Arcor2Exception


@dataclass
class MeshFocusAction(JsonSchemaMixin):

    mesh_focus_points: list[Position]
    robot_space_points: list[Position]

    def __post_init__(self) -> None:

        if len(self.mesh_focus_points) != len(self.robot_space_points):
            raise Arcor2Exception("Invalid MeshFocusAction.")


@dataclass
class LineCheck(JsonSchemaMixin):

    pt1: Position
    pt2: Position


@dataclass
class LineCheckResult(JsonSchemaMixin):

    safe: bool
    object_id: Optional[str] = None



import importlib
import inspect
import pkgutil
from typing import TYPE_CHECKING

import dataclasses_jsonschema
import orjson
from dataclasses_jsonschema import JsonSchemaMixin, ValidationError
from dataclasses_jsonschema.apispec import DataclassesPlugin, _schema_reference

from arcor2.exceptions import Arcor2Exception


class DataException(Arcor2Exception):
    pass


if not TYPE_CHECKING:

    # monkey-patch dataclasses_jsonschema to use orjson instead of json

    @classmethod
    def from_json(
        cls: type[dataclasses_jsonschema.T], data: str, validate: bool = True, **json_kwargs
    ) -> dataclasses_jsonschema.T:
        return cls.from_dict(orjson.loads(data), validate)

    def to_json(self, omit_none: bool = True, validate: bool = False, **json_kwargs) -> str:
        return orjson.dumps(self.to_dict(omit_none, validate), **json_kwargs).decode()

    JsonSchemaMixin.from_json = from_json
    JsonSchemaMixin.to_json = to_json


def resolve_schema_refs(self, data: dict) -> None:

    if "schema" in data:

        if "$ref" in data["schema"]:
            data["schema"]["$ref"] = _schema_reference(data["schema"]["$ref"], self._schema_type)
        elif "items" in data["schema"] and "$ref" in data["schema"]["items"]:
            data["schema"]["items"]["$ref"] = _schema_reference(data["schema"]["items"]["$ref"], self._schema_type)
    else:
        for key in data:
            if isinstance(data[key], dict):
                self.resolve_schema_refs(data[key])


# monkey patch to solve https://github.com/s-knibbs/dataclasses-jsonschema/issues/126
DataclassesPlugin.resolve_schema_refs = resolve_schema_refs  # type: ignore


def compile_json_schemas() -> None:
    """
    Force compilation of json schema (otherwise it might cause troubles later when executed in parallel)
    :return:
    """

    from arcor2 import data

    for _, module_name, _ in pkgutil.iter_modules(data.__path__):

        module = importlib.import_module(f"{data.__name__}.{module_name}")

        for _, obj in inspect.getmembers(module, inspect.isclass):

            if not issubclass(obj, JsonSchemaMixin) or obj is JsonSchemaMixin or inspect.isabstract(obj):
                continue

            try:
                obj.from_dict({})
            except ValidationError:
                pass



import os
from datetime import datetime

from dateutil.parser import parse

from arcor2 import rest
from arcor2.data.common import IdDesc, Project, ProjectParameter, ProjectSources, Scene, SceneObjectOverride
from arcor2.data.object_type import MODEL_MAPPING, Mesh, MeshList, MetaModel3d, Model, Model3dType, ObjectType
from arcor2.exceptions import Arcor2Exception
from arcor2.exceptions.helpers import handle
from arcor2.logging import get_logger

URL = os.getenv("ARCOR2_PROJECT_SERVICE_URL", "http://0.0.0.0:10000")

"""
Collection of functions to work with the Project service (0.10.0).

All functions raise ProjectServiceException when any failure happens.

"""

logger = get_logger("Project")


class ProjectServiceException(Arcor2Exception):
    pass


# ----------------------------------------------------------------------------------------------------------------------
# Models
# ----------------------------------------------------------------------------------------------------------------------


@handle(ProjectServiceException, logger, message="Failed to list models.")
def get_models() -> list[MetaModel3d]:
    return rest.call(rest.Method.GET, f"{URL}/models", list_return_type=MetaModel3d)


@handle(ProjectServiceException, logger, message="Failed to get the mesh.")
def get_mesh(mesh_id: str) -> Mesh:
    return rest.call(rest.Method.GET, f"{URL}/models/{mesh_id}/mesh", return_type=Mesh)


@handle(ProjectServiceException, logger, message="Failed to get list of meshes.")
def get_meshes() -> MeshList:
    return rest.call(rest.Method.GET, f"{URL}/models/meshes", list_return_type=Mesh)


@handle(ProjectServiceException, logger, message="Failed to get the model type.")
def get_model(model_id: str, model_type: Model3dType) -> Model:
    return rest.call(
        rest.Method.GET, f"{URL}/models/{model_id}/{model_type.value.lower()}", return_type=MODEL_MAPPING[model_type]
    )


@handle(ProjectServiceException, logger, message="Failed to add or update the model.")
def put_model(model: Model) -> None:
    rest.call(rest.Method.PUT, f"{URL}/models/{model.__class__.__name__.lower()}", body=model)


@handle(ProjectServiceException, logger, message="Failed to delete the model.")
def delete_model(model_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/models/{model_id}")


# ----------------------------------------------------------------------------------------------------------------------
# ObjectParameters
# ----------------------------------------------------------------------------------------------------------------------


@handle(ProjectServiceException, logger, message="Failed to get project parameters.")
def get_object_parameters(project_id: str) -> list[SceneObjectOverride]:
    return rest.call(
        rest.Method.GET, f"{URL}/projects/{project_id}/object-parameters", list_return_type=SceneObjectOverride
    )


@handle(ProjectServiceException, logger, message="Failed to add or update project parameters.")
def update_object_parameters(project_id: str, parameters: list[SceneObjectOverride]) -> None:
    rest.call(rest.Method.GET, f"{URL}/projects/{project_id}/object-parameters", body=parameters)


# ----------------------------------------------------------------------------------------------------------------------
# ObjectTypes
# ----------------------------------------------------------------------------------------------------------------------


@handle(ProjectServiceException, logger, message="Failed to get the object type.")
def get_object_type(object_type_id: str) -> ObjectType:
    obj_type = rest.call(rest.Method.GET, f"{URL}/object-types/{object_type_id}", return_type=ObjectType)
    assert obj_type.modified, f"Project service returned object without 'modified': {obj_type.id}."
    return obj_type


@handle(ProjectServiceException, logger, message="Failed to list object types.")
def get_object_type_ids() -> list[IdDesc]:
    return rest.call(rest.Method.GET, f"{URL}/object-types", list_return_type=IdDesc)


@handle(ProjectServiceException, logger, message="Failed to add or update the object type.")
def update_object_type(object_type: ObjectType) -> datetime:

    assert object_type.id
    return parse(rest.call(rest.Method.PUT, f"{URL}/object-types", body=object_type, return_type=str))


@handle(ProjectServiceException, logger, message="Failed to delete the object type.")
def delete_object_type(object_type_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/object-types/{object_type_id}")


# ----------------------------------------------------------------------------------------------------------------------
# Parameters
# ----------------------------------------------------------------------------------------------------------------------


@handle(ProjectServiceException, logger, message="Failed to get project parameters.")
def get_project_parameters(project_id: str) -> list[ProjectParameter]:
    return rest.call(rest.Method.GET, f"{URL}/projects/{project_id}/parameters", list_return_type=ProjectParameter)


@handle(ProjectServiceException, logger, message="Failed to add or update project parameters.")
def update_project_parameters(project_id: str, parameters: list[ProjectParameter]) -> datetime:
    return parse(
        rest.call(rest.Method.GET, f"{URL}/projects/{project_id}/parameters", body=parameters, return_type=str)
    )


# ----------------------------------------------------------------------------------------------------------------------
# Projects
# ----------------------------------------------------------------------------------------------------------------------


@handle(ProjectServiceException, logger, message="Failed to list projects.")
def get_projects() -> list[IdDesc]:
    return rest.call(rest.Method.GET, f"{URL}/projects", list_return_type=IdDesc)


@handle(ProjectServiceException, logger, message="Failed to get the project.")
def get_project(project_id: str) -> Project:
    return rest.call(rest.Method.GET, f"{URL}/projects/{project_id}", return_type=Project)


@handle(ProjectServiceException, logger, message="Failed to get the project sources.")
def get_project_sources(project_id: str) -> ProjectSources:
    return rest.call(rest.Method.GET, f"{URL}/projects/{project_id}/sources", return_type=ProjectSources)


@handle(ProjectServiceException, logger, message="Failed to add or update the project.")
def update_project(project: Project) -> datetime:

    assert project.id
    return parse(rest.call(rest.Method.PUT, f"{URL}/projects", return_type=str, body=project))


@handle(ProjectServiceException, logger, message="Failed to add or update the project sources.")
def update_project_sources(project_sources: ProjectSources) -> None:

    assert project_sources.id
    rest.call(rest.Method.PUT, f"{URL}/projects/sources", body=project_sources)


@handle(ProjectServiceException, logger, message="Failed to delete the project.")
def delete_project(project_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/projects/{project_id}")


@handle(ProjectServiceException, logger, message="Failed to clone the project.")
def clone_project(
    project_id: str, new_project_name: str, new_description: None | str = None, new_project_id: None | str = None
) -> Project:

    if not new_project_id:
        new_project_id = Project.uid()

    params: dict[str, str] = {
        "project_id": project_id,
        "new_project_name": new_project_name,
        "new_project_id": new_project_id,
    }

    if new_description:
        params["new_description"] = new_description

    return rest.call(rest.Method.PUT, f"{URL}/projects/clone", params=params, return_type=Project)


# ----------------------------------------------------------------------------------------------------------------------
# Scenes
# ----------------------------------------------------------------------------------------------------------------------


@handle(ProjectServiceException, logger, message="Failed to list scenes.")
def get_scenes() -> list[IdDesc]:
    return rest.call(rest.Method.GET, f"{URL}/scenes", list_return_type=IdDesc)


@handle(ProjectServiceException, logger, message="Failed to get the scene.")
def get_scene(scene_id: str) -> Scene:
    return rest.call(rest.Method.GET, f"{URL}/scenes/{scene_id}", return_type=Scene)


@handle(ProjectServiceException, logger, message="Failed to add or update the scene.")
def update_scene(scene: Scene) -> datetime:

    assert scene.id
    return parse(rest.call(rest.Method.PUT, f"{URL}/scenes", return_type=str, body=scene))


@handle(ProjectServiceException, logger, message="Failed to delete the scene.")
def delete_scene(scene_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/scenes/{scene_id}")



import io

import cv2
import numpy as np
import PIL.Image
from PIL.Image import Image

from arcor2 import json

ENCODING = "latin-1"


def image_to_cv2(pil_image: Image, mode=cv2.COLOR_RGB2BGR) -> np.ndarray:

    return cv2.cvtColor(np.array(pil_image), mode)


def image_to_bytes_io(value: Image, target_format: str = "jpeg", target_mode: None | str = None) -> io.BytesIO:

    output = io.BytesIO()

    if target_mode and value.mode != target_mode:
        rgb_im = value.convert(target_mode)
        rgb_im.save(output, target_format)
    else:
        value.save(output, target_format)
    output.seek(0)
    return output


def image_to_str(value: Image, target_format: str = "jpeg") -> str:
    return image_to_bytes_io(value, target_format).getvalue().decode(ENCODING)


def image_from_str(value: str) -> Image:
    return image_from_bytes_io(io.BytesIO(value.encode(ENCODING)))


def image_to_json(value: Image) -> str:
    return json.dumps(image_to_str(value))


def image_from_bytes_io(value: io.BytesIO) -> Image:
    value.seek(0)
    return PIL.Image.open(value)


def image_from_json(value: str) -> Image:
    return image_from_str(json.loads_type(value, str))



import copy
from ast import Attribute, Load, Name
from typing import Any

from arcor2 import json
from arcor2 import transformations as tr
from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import Pose
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins import ParameterPluginException
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class PosePlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return Pose

    @classmethod
    def orientation_id(cls, project: CProject, action_id: str, parameter_id: str) -> str:
        return cls._id_from_value(project.action(action_id).parameter(parameter_id).value)

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Pose:

        try:
            ap, ori = project.bare_ap_and_orientation(cls.orientation_id(project, action_id, parameter_id))
        except Arcor2Exception as e:
            raise ParameterPluginException("Failed to get scene/project data.") from e
        return Pose(ap.position, ori.orientation)

    @classmethod
    def parameter_execution_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Pose:

        # return copy in order to avoid unwanted changes in the original value if an action modifies the parameter
        return copy.deepcopy(
            tr.abs_pose_from_ap_orientation(scene, project, cls.orientation_id(project, action_id, parameter_id))
        )

    @classmethod
    def value_to_json(cls, value: Pose) -> str:
        return value.to_json()

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:

        return orientation_id == cls.orientation_id(project, action_id, parameter_id)

    @classmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Attribute:

        ori_ap, ori = project.bare_ap_and_orientation(cls.orientation_id(project, action_id, parameter_id))

        return Attribute(
            value=Attribute(
                value=Attribute(
                    value=Name(id="aps", ctx=Load()), attr=ori_ap.name, ctx=Load()  # TODO this should not be hardcoded
                ),
                attr="poses",  # TODO this should not be hardcoded
                ctx=Load(),
            ),
            attr=ori.name,
            ctx=Load(),
        )


class PoseListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return list[Pose]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(PosePlugin)

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> list[Pose]:

        ret: list[Pose] = []

        ap, action = project.action_point_and_action(action_id)
        parameter = action.parameter(parameter_id)

        for orientation_id in cls._param_value_list(parameter):
            ret.append(Pose(ap.position, project.orientation(orientation_id).orientation))

        return ret

    @classmethod
    def parameter_execution_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> list[Pose]:

        ap, action = project.action_point_and_action(action_id)

        if not ap.parent:
            return copy.deepcopy(cls.parameter_value(type_defs, scene, project, action_id, parameter_id))

        parameter = action.parameter(parameter_id)
        ret: list[Pose] = []

        for orientation_id in cls._param_value_list(parameter):
            ret.append(copy.deepcopy(tr.abs_pose_from_ap_orientation(scene, project, orientation_id)))

        return ret

    @classmethod
    def value_to_json(cls, value: list[Pose]) -> str:
        return json.dumps([v.to_json() for v in value])

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:

        action = project.action(action_id)
        param = action.parameter(parameter_id)

        for ori_id in cls._param_value_list(param):
            if ori_id == orientation_id:
                return True
        return False



from ast import Str
from typing import Any

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class StringPlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return str

    @classmethod
    def type_name(cls) -> str:
        return "string"

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> str:
        return super(StringPlugin, cls).parameter_value(type_defs, scene, project, action_id, parameter_id)

    @classmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Str:
        return Str(s=cls.parameter_execution_value(type_defs, scene, project, action_id, parameter_id), kind="")


class StringListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return list[str]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(StringPlugin)

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> list[str]:
        return super(StringListPlugin, cls).parameter_value(type_defs, scene, project, action_id, parameter_id)



import random

from arcor2.data.common import ActionMetadata
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic


class RandomActions(Generic):
    """Collection of actions to generate random values."""

    _ABSTRACT = False

    def random_integer(self, range_min: int, range_max: int, *, an: None | str = None) -> int:
        """Generates random integer in given range (including min/max values).

        :param range_min: Minimal value.
        :param range_max: Maximal value.
        :return:
        """

        try:
            return random.randint(range_min, range_max)
        except ValueError as e:
            raise Arcor2Exception(str(e)) from e

    def random_double(self, range_min: float, range_max: float, *, an: None | str = None) -> float:
        """Generates random double in given range.

        :param range_min: Minimal value.
        :param range_max: Maximal value.
        :return:
        """

        try:
            return random.uniform(range_min, range_max)
        except ValueError as e:
            raise Arcor2Exception(str(e)) from e

    def random_bool(self, *, an: None | str = None) -> bool:
        """Returns random boolean value.

        :return:
        """
        return random.choice((False, True))

    random_integer.__action__ = ActionMetadata()  # type: ignore
    random_double.__action__ = ActionMetadata()  # type: ignore
    random_bool.__action__ = ActionMetadata()  # type: ignore



from typing import Any, Sequence, TypeVar

import orjson

from arcor2.exceptions import Arcor2Exception


class JsonException(Arcor2Exception):
    pass


JsonType = bool | str | int | float | dict[str, Any] | Sequence[Any] | None


T = TypeVar("T")


def loads(value: str) -> JsonType:

    try:
        return orjson.loads(value)
    except (ValueError, TypeError) as e:
        raise JsonException(f"Not a JSON. {str(e)}") from e


def loads_type(value: str, output_type: type[T]) -> T:

    val = loads(value)

    if not isinstance(val, output_type):
        raise JsonException("Not an expected type.")

    return val


def dumps(value: JsonType) -> str:

    try:
        return orjson.dumps(value).decode()
    except (ValueError, TypeError) as e:
        raise JsonException(f"Not a JSON. {str(e)}") from e



import os
import zipfile
from io import BytesIO
from typing import NamedTuple

from arcor2.clients import asset
from arcor2.clients import project_service as ps
from arcor2.data.object_type import Mesh, Models, ObjectType
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import CollisionObject, Generic, Robot
from arcor2.object_types.utils import check_object_type, get_containing_module_sources


class UploadException(Arcor2Exception):
    pass


class Urdf(NamedTuple):

    path_to_directory: str  # path to a URDF package directory
    archive_name: str  # resulting archive name


def upload_whatever(type_def: type[object]) -> None:

    obj_type = ObjectType(id=type_def.__name__, source=get_containing_module_sources(type_def))
    print(f"Storing '{obj_type.id}'...")
    ps.update_object_type(obj_type)


def upload_def(
    type_def: type[Generic],
    model: None | Models = None,
    urdf: None | Urdf = None,
    file_to_upload: None | str = None,
) -> None:
    """Uploads ObjectType definition to the Project service.

    :param type_def: Class definition.
    :param model: Collision model.
    :param urdf: If the type is robot, path to its URDF can be given here.
    :param file_to_upload: Path to a file. Used e.g. to upload a mesh together with the ObjectType.
    :return:
    """

    if not issubclass(type_def, Robot) and urdf:
        raise UploadException("Parameter 'urdf' set for non-Robot.")

    try:
        check_object_type(type_def)
    except Arcor2Exception as e:
        raise UploadException(f"{type_def.__name__} is not a valid ObjectType. {str(e)}")

    obj_type = ObjectType(
        id=type_def.__name__, source=get_containing_module_sources(type_def), description=type_def.description()
    )

    if issubclass(type_def, CollisionObject):
        if not model:
            raise UploadException("Parameter 'model' must be set for CollisionObject.")

        if model.id != obj_type.id:
            raise UploadException("Model id have to be the same as ObjectType id.")

        obj_type.model = model.metamodel()

        if isinstance(model, Mesh):

            if not type_def.mesh_filename:
                raise UploadException("Mesh filename not set.")

            if not file_to_upload:
                raise UploadException("For mesh collision model, file_to_upload parameter have to be set.")

            try:
                with open(file_to_upload, "rb") as f:
                    asset.create_asset(
                        type_def.mesh_filename,
                        f.read(),
                        directory_path="/meshes",
                        description=f"mesh for {obj_type.id}",
                    )
            except OSError as e:
                raise UploadException(f"Failed to read mesh file. {str(e)}")

        ps.put_model(model)

    else:
        if model:
            raise UploadException("Parameter 'model' set for non-CollisionObject.")

    if urdf:

        if not os.path.isdir(urdf.path_to_directory):
            print(f"{urdf.path_to_directory} is not a directory.")
            return

        print(f"Storing URDF package for '{obj_type.id}'.")

        paths: list[str] = []

        # full path is in form src/python/arcor2_fit_demo/data/dobot-m1/dobot_m1_description/meshes
        # so we need to remove the prefix (e.g. src/python/arcor2_fit_demo/data/dobot-m1)
        # let's find the prefix...
        for root, _, files in os.walk(urdf.path_to_directory):
            for filename in files:
                paths.append(os.path.join(root, filename))

        prefix = os.path.commonpath(paths)

        mem_zip = BytesIO()
        with zipfile.ZipFile(mem_zip, mode="w", compression=zipfile.ZIP_DEFLATED) as zf:
            for root, _, files in os.walk(urdf.path_to_directory):
                for filename in files:
                    path = os.path.join(root, filename)
                    # in the archive, the path will be without the prefix
                    zf.write(path, os.path.relpath(path, prefix))

        mem_zip.seek(0)
        asset.create_asset(
            urdf.archive_name,
            mem_zip.getvalue(),
            directory_path="/urdf_packages",
            description=f"URDF for {obj_type.id}",
        )

    print(f"Storing '{obj_type.id}'...")
    ps.update_object_type(obj_type)



import asyncio
import builtins
import importlib
import keyword
import os
import socket
import sys
from concurrent import futures
from contextlib import closing
from threading import Lock
from typing import Any, Callable, TypeVar

import humps
from packaging.version import Version, parse

from arcor2.exceptions import Arcor2Exception


class ImportClsException(Arcor2Exception):
    pass


class TypeDefException(Arcor2Exception):
    pass


def is_valid_identifier(value: str) -> None:
    """
    Identifier (e.g. object id) will be used as variable name in the script - it should
    not contain any special characters etc.
    :param value:
    :return:
    """

    if not value:
        raise Arcor2Exception("Empty string is not enough.")

    if not (value[0].isalpha() or value[0] == "_"):
        raise Arcor2Exception("It has to start with character or underscore.")

    for c in value[1:]:
        if c == " ":
            raise Arcor2Exception("Use underscore instead of space.")
        if not (c.isalnum() or c == "_"):
            raise Arcor2Exception(f"Character '{c}' can't be used.")

    assert value.isidentifier()

    if keyword.iskeyword(value) or value in builtins.__dict__.keys():
        raise Arcor2Exception("Reserved keyword.")


def is_valid_type(value: str) -> None:
    is_valid_identifier(value)


S = TypeVar("S")


async def run_in_executor(
    func: Callable[..., S],
    *args: Any,
    executor: None | futures.Executor = None,
    propagate: None | list[type[Exception]] = None,
) -> S:
    """Executes synchronous function in an executor. Catches all exceptions are
    re-raises them as Arcor2Exception.

    :param func:
    :param args:
    :param executor:
    :param propagate: Exceptions to propagate.
    :return:
    """

    # TODO user typing.ParamSpec instead of *args: Any (Python 3.10 or typing-extensions)
    # ...not supported by mypy at the moment, see https://github.com/python/mypy/issues/8645

    try:
        return await asyncio.get_event_loop().run_in_executor(executor, func, *args)
    except Arcor2Exception:
        raise
    except Exception as e:

        if propagate:
            for etp in propagate:
                if isinstance(e, etp):
                    raise

        # all code should raise exceptions based on Arcor2Exception so this is just a guard against a buggy code
        raise Arcor2Exception(f"Unhandled exception in {func.__name__}.") from e


T = TypeVar("T")


def save_and_import_type_def(source: str, type_name: str, output_type: type[T], path: str, module_name: str) -> type[T]:
    """Save source to a file in object_type_path directory.

    :param source:
    :param type_name:
    :param output_type:
    :param object_type_path:
    :return:
    """

    type_file = humps.depascalize(type_name)
    full_path = f"{os.path.join(path, module_name, type_file)}.py"

    with open(full_path, "w") as file:
        file.write(source)

    try:
        return import_type_def(type_name, output_type, path, module_name)
    except Arcor2Exception:
        os.remove(full_path)
        raise


def import_type_def(type_name: str, output_type: type[T], path: str, module_name: str) -> type[T]:
    """Save source to a file in object_type_path directory.

    :param source:
    :param type_name:
    :param output_type:
    :param object_type_path:
    :return:
    """

    if path not in sys.path:
        sys.path.append(path)

    type_file = humps.depascalize(type_name)

    importlib.invalidate_caches()  # otherwise import might fail randomly (not sure why exactly)

    try:
        module = importlib.import_module(f"{module_name}.{type_file}")

        # reload is necessary for cases when the module is already loaded
        module = importlib.reload(module)  # TODO does this really solve anything?

    except ImportError as e:
        raise ImportClsException(f"Failed to import '{module_name}.{type_file}'. {str(e).capitalize()}.") from e
    except Exception as e:
        raise ImportClsException(
            f"Error occurred when importing '{module_name}.{type_file}'. {str(e).capitalize()}."
        ) from e

    if not module.__file__:
        raise ImportClsException(f"Can't get file for '{module_name}.{type_file}'.")

    path_to_file = os.path.abspath(module.__file__)
    assert os.path.exists(path_to_file), f"Path {path_to_file} does not exist."

    try:
        cls = getattr(module, type_name)
    except AttributeError:
        raise ImportClsException(f"Class {type_name} not found in module '{module_name}'.")

    if not issubclass(cls, output_type):
        raise ImportClsException(f"{cls.__name__} is not subclass of {output_type.__name__}.")

    return cls


def check_compatibility(my_version: str, their_version: str) -> None:

    try:
        mv = parse(my_version)
        tv = parse(their_version)
    except ValueError as e:
        raise Arcor2Exception from e

    if not isinstance(mv, Version):
        raise Arcor2Exception(f"{my_version} is not a proper version.")

    if not isinstance(tv, Version):
        raise Arcor2Exception(f"{their_version} is not a proper version.")

    if mv.major != tv.major:
        raise Arcor2Exception("Different major version.")

    if mv.major == 0:
        if mv.minor != tv.minor:
            raise Arcor2Exception(f"Our version {my_version} is not compatible with {their_version}.")
    else:
        if mv.minor > tv.minor:
            raise Arcor2Exception(f"Our version {my_version} is outdated for {their_version}.")


class NonBlockingLock:
    """This lock can only be used as a context manager."""

    __slots__ = ("_lock",)

    def __init__(self) -> None:
        self._lock = Lock()

    def locked(self) -> bool:
        return self._lock.locked()

    def __enter__(self) -> None:
        if not self._lock.acquire(blocking=False):
            raise Arcor2Exception("Can't get the lock.")

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._lock.release()
        return None


def port_from_url(url: str) -> int:

    return int(url.strip().split(":")[-1])


def find_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("localhost", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]



from typing import NamedTuple

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedProjectException
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import BareActionPoint, Orientation, Pose, Position
from arcor2.exceptions import Arcor2Exception


def make_pose_rel(parent: Pose, child: Pose) -> Pose:
    """
    :param parent: e.g. scene object
    :param child:  e.g. action point
    :return: relative pose
    """

    return Pose(
        (child.position - parent.position).rotated(parent.orientation, inverse=True),
        Orientation.from_quaternion(parent.orientation.as_quaternion().inverse() * child.orientation.as_quaternion()),
    )


def make_pose_abs(parent: Pose, child: Pose) -> Pose:
    """
    :param parent: e.g. scene object
    :param child:  e.g. action point
    :return: absolute pose
    """

    return Pose(
        (child.position.rotated(parent.orientation) + parent.position),
        Orientation.from_quaternion(parent.orientation.as_quaternion() * child.orientation.as_quaternion()),
    )


class Parent(NamedTuple):

    pose: Pose
    parent_id: None | str = None  # parent of the parent


def get_parent_pose(scene: CScene, project: CProject, parent_id: str) -> Parent:
    """Returns pose of the parent and parent of the parent (if any).

    :param scene:
    :param project:
    :param parent_id:
    :return:
    """

    if parent_id in scene.object_ids:
        parent_obj = scene.object(parent_id)
        if not parent_obj.pose:
            raise Arcor2Exception("Parent object does not have pose!")
        # TODO find parent object in the graph (SceneObject has "children" property)
        return Parent(parent_obj.pose, None)
    elif parent_id in project.action_points_ids:
        ap = project.bare_action_point(parent_id)
        return Parent(Pose(ap.position, Orientation()), ap.parent)
    else:
        raise Arcor2Exception(f"Unknown parent_id {parent_id}.")


def make_relative_ap_global(scene: CScene, project: CProject, ap: BareActionPoint) -> set[str]:
    """Transforms (in place) relative AP into a global one.

    :param scene:
    :param project:
    :param ap:
    :return:
    """

    if not ap.parent:
        raise Arcor2Exception("Action point does not have a parent.")

    updated_aps: set[str] = set()

    def _update_childs(parent: Parent, ap_id: str) -> None:

        for child_id in project.childs(ap_id):

            try:
                child_ap = project.bare_action_point(child_id)
            except CachedProjectException:
                continue

            child_ap.position = child_ap.position.rotated(parent.pose.orientation)

            for ori in project.ap_orientations(child_ap.id):
                ori.orientation = Orientation.from_quaternion(
                    parent.pose.orientation.as_quaternion() * ori.orientation.as_quaternion()
                )

            updated_aps.add(child_ap.id)
            _update_childs(parent, child_ap.id)

    def _make_relative_ap_global(_ap: BareActionPoint) -> None:

        if not _ap.parent:
            return

        parent = get_parent_pose(scene, project, _ap.parent)

        if parent.pose.orientation != Orientation():
            _update_childs(parent, _ap.id)

        _ap.position = make_pose_abs(parent.pose, Pose(_ap.position, Orientation())).position
        for ori in project.ap_orientations(_ap.id):
            ori.orientation = make_pose_abs(parent.pose, Pose(Position(), ori.orientation)).orientation

        _ap.parent = parent.parent_id
        _make_relative_ap_global(_ap)

    _make_relative_ap_global(ap)
    return updated_aps


def make_global_ap_relative(scene: CScene, project: CProject, ap: BareActionPoint, parent_id: str) -> set[str]:
    """Transforms (in place) global AP into a relative one with given parent
    (can be object or another AP).

    :param scene:
    :param project:
    :param ap:
    :param parent_id:
    :return:
    """

    assert project.scene_id == scene.id

    if ap.parent:
        raise Arcor2Exception("Action point already has a parent.")

    updated_aps: set[str] = set()
    updated_orientations: set[str] = set()

    def _update_childs(parent: Parent, ap_id: str) -> None:

        for child_id in project.childs(ap_id):

            try:
                child_ap = project.bare_action_point(child_id)
            except CachedProjectException:
                continue

            child_ap.position = child_ap.position.rotated(parent.pose.orientation, inverse=True)

            for ori in project.ap_orientations(child_ap.id):
                ori.orientation = Orientation.from_quaternion(
                    parent.pose.orientation.as_quaternion().inverse() * ori.orientation.as_quaternion()
                )

                updated_orientations.add(ori.id)

            updated_aps.add(child_ap.id)
            _update_childs(parent, child_ap.id)

    def _make_global_ap_relative(parent_id: str) -> None:

        parent = get_parent_pose(scene, project, parent_id)

        if parent.parent_id:
            _make_global_ap_relative(parent.parent_id)

        ap.position = make_pose_rel(parent.pose, Pose(ap.position, Orientation())).position
        for ori in project.ap_orientations(ap.id):
            ori.orientation = make_pose_rel(parent.pose, Pose(Position(), ori.orientation)).orientation

        if parent.pose.orientation != Orientation():
            _update_childs(parent, ap.id)

    _make_global_ap_relative(parent_id)
    ap.parent = parent_id
    return updated_aps


def make_pose_rel_to_parent(scene: CScene, project: CProject, pose: Pose, parent_id: str) -> Pose:
    """Transforms global Pose into Pose that is relative to a given parent (can
    be object or AP).

    :param scene:
    :param project:
    :param pose:
    :param parent_id:
    :return:
    """

    parent = get_parent_pose(scene, project, parent_id)
    if parent.parent_id:
        pose = make_pose_rel_to_parent(scene, project, pose, parent.parent_id)

    return make_pose_rel(parent.pose, pose)


def abs_pose_from_ap_orientation(scene: CScene, project: CProject, orientation_id: str) -> Pose:
    """Returns absolute Pose without modifying anything within the project.

    :param orientation_id:
    :return:
    """

    ap, ori = project.bare_ap_and_orientation(orientation_id)

    pose = Pose(ap.position, ori.orientation)
    parent_id = ap.parent

    while parent_id:
        parent = get_parent_pose(scene, project, parent_id)
        pose = make_pose_abs(parent.pose, pose)
        parent_id = parent.parent_id

    return pose


def abs_position_from_ap(scene: CScene, project: CProject, ap_id: str) -> Position:
    """Returns absolute Position without modifying anything within the project.

    :param ap_id:
    :return:
    """

    ap = project.bare_action_point(ap_id)
    pose = Pose(ap.position, Orientation())
    parent_id = ap.parent

    while parent_id:
        parent = get_parent_pose(scene, project, parent_id)
        pose = make_pose_abs(parent.pose, pose)
        parent_id = parent.parent_id

    return pose.position



import pkgutil
import warnings
from typing import NamedTuple


class DynamicParamTuple(NamedTuple):

    method_name: str
    parent_parameters: set[str]


# key: name of parameter, value: name of method to call (to get set of strings), set of parent parameters
DynamicParamDict = dict[str, DynamicParamTuple]

# key: action, value: cancel method
CancelDict = dict[str, str]


if __debug__:
    warnings.filterwarnings("error", module="dataclasses_jsonschema")  # make warnings fatal because they are


def package_version(package: str, file: str = "VERSION") -> str:

    if res := pkgutil.get_data(package, file):
        return res.decode().strip()
    return "unknown"


def version() -> str:
    return package_version(__name__)



import time

from arcor2.data.common import ActionMetadata, Position
from arcor2.object_types.abstract import Generic


class PositionParam(Generic):

    _ABSTRACT = False

    def action_with_position_param(self, param: Position, *, an: None | str = None) -> None:
        """Pointless action.

        :param param:
        :return:
        """
        time.sleep(0.1)

    action_with_position_param.__action__ = ActionMetadata()  # type: ignore



from arcor2.cached import CachedProject, CachedProjectException
from arcor2.data.common import Action
from arcor2.exceptions import Arcor2Exception

LogicContainer = CachedProject  # TODO make it Union[CachedProject, CachedProjectFunction]

# TODO def validate_logic() -> to be called before saving project / building the package


def check_for_loops(parent: LogicContainer, first_action_id: None | str = None) -> None:
    """Finds loops in logic. Can process even unfinished logic, when first
    action id is provided.

    :param parent:
    :param first_action_id:
    :return:
    """

    def _check_for_loops(action: Action, visited_actions: set[str]) -> None:

        if action.id in visited_actions:
            raise Arcor2Exception("Loop detected!")

        visited_actions.add(action.id)

        _, outputs = parent.action_io(action.id)

        for output in outputs:

            if output.end == output.END:
                continue

            # each possible execution path have its own set of visited actions
            _check_for_loops(parent.action(output.end), visited_actions.copy())

    if first_action_id is None:

        try:
            first_action_id = parent.first_action_id()
        except CachedProjectException as e:
            raise Arcor2Exception("Can't check unfinished logic.") from e

    first_action = parent.action(first_action_id)

    visited_actions: set[str] = set()
    _check_for_loops(first_action, visited_actions)



import os
from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2 import rest

URL = os.getenv("ARCOR2_ASSET_SERVICE_URL", "http://0.0.0.0:10040")


@dataclass
class AssetInfo(JsonSchemaMixin):

    id: str
    type: str
    created: datetime
    modified: datetime
    file_path: Optional[str] = None
    file_name: Optional[str] = None
    name: Optional[str] = None
    extension: Optional[str] = None
    directory_path: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[list[str]] = None
    dependencies: Optional[list[str]] = None


def asset_info() -> list[AssetInfo]:

    return rest.call(rest.Method.GET, f"{URL}/assets/info", list_return_type=AssetInfo)


def asset_ids() -> set[str]:
    return {ai.id for ai in asset_info()}


def create_asset(
    id: str,  # on Asset service, id is not required but for arcor2, it makes more sense to require it
    asset_data: bytes,
    *,
    directory_path: str | None = None,
    description: str | None = None,
    type: str | None = None,
    dependencies: list[str] | None = None,
    tags: list[str] | None = None,
    upsert: bool = True,
) -> AssetInfo:
    """Creates the asset."""

    if upsert and asset_exists(id):
        delete_asset(id, remove_dependers=True)

    return rest.call(
        rest.Method.POST,
        f"{URL}/assets",
        params={
            "id": id,
            "directory_path": directory_path,
            "description": description,
            "type": type,
            "dependencies": dependencies,
            "tags": tags,
        },
        files={"asset_data": (id, asset_data)},
        return_type=AssetInfo,
    )


def delete_asset(id: str, remove_dependers: bool = False) -> None:

    rest.call(rest.Method.DELETE, f"{URL}/assets/{id}", params={"remove_dependers": remove_dependers})


def asset_exists(id: str) -> bool:

    return rest.call(rest.Method.GET, f"{URL}/assets/{id}/exists", return_type=bool)



import logging

import aiologger
import colorlog


def getMessage(self):
    """backwards compatibility for python builtin logger ref.

    python3.7/logging/__init__.py", line 608
    """
    return self.get_message()


# see https://github.com/b2wdigital/aiologger/pull/85
aiologger.records.LogRecord.getMessage = getMessage


DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
DEFAULT_LOG_FORMAT = "%(asctime)s %(log_color)s%(levelname)-8s%(reset)s %(message)s"


def logger_formatter(log_format=DEFAULT_LOG_FORMAT, date_format=DEFAULT_DATE_FORMAT) -> logging.Formatter:

    return colorlog.ColoredFormatter(log_format, date_format)


def get_aiologger(
    name: str,
    level: aiologger.levels.LogLevel = aiologger.levels.LogLevel.INFO,
    formatter: None | logging.Formatter = None,
) -> aiologger.Logger:

    if formatter is None:
        formatter = logger_formatter()

    return aiologger.Logger.with_default_handlers(name=name, formatter=formatter, level=level)


def get_logger(name: str, level: int = logging.INFO) -> logging.Logger:

    logger = logging.getLogger(name)

    ch = logging.StreamHandler()
    ch.setFormatter(logger_formatter())
    logger.setLevel(level)
    logger.addHandler(ch)

    return logger



from arcor2.data.common import ActionMetadata
from arcor2.object_types.abstract import Generic, Settings


class FlowActions(Generic):
    """Provides actions related to program flow."""

    _ABSTRACT = False

    def __init__(self, obj_id: str, name: str, settings: None | Settings = None) -> None:
        super().__init__(obj_id, name, settings)
        self._first_pass = True

    def is_first_pass(self, *, an: None | str = None) -> bool:
        """Returns True for the first pass of the main cycle.

        :param an:
        :return:
        """

        if ret := self._first_pass:
            self._first_pass = False

        return ret

    is_first_pass.__action__ = ActionMetadata()  # type: ignore



import abc
import ast
import copy
from ast import AST
from typing import Any, Callable, NamedTuple

import humps

from arcor2 import json
from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.object_type import ParameterMeta
from arcor2.exceptions import Arcor2NotImplemented
from arcor2.parameter_plugins import ParameterPluginException, TypesDict


class ImportTuple(NamedTuple):

    module_name: str
    class_name: str


class ParameterPlugin(metaclass=abc.ABCMeta):

    EXACT_TYPE = True
    COUNTABLE = False

    @classmethod
    def _id_from_value(cls, value: str) -> str:

        arbitrary_id = cls._value_from_json(value)

        if not isinstance(arbitrary_id, str):
            raise ParameterPluginException("String expected.")

        return arbitrary_id

    @classmethod
    @abc.abstractmethod
    def type(cls) -> Any:
        """Returns python type."""
        raise Arcor2NotImplemented()

    @classmethod
    def type_name(cls) -> str:
        """Returns parameter type as string used in JSON."""
        return humps.depascalize(cls.type().__name__)

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:

        assert param_meta.type == cls.type_name()
        assert param_meta.name

    @classmethod
    @abc.abstractmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Any:

        action = project.action(action_id)
        param_value = action.parameter(parameter_id).value

        try:
            val = cls._value_from_json(param_value)
        except ParameterPluginException as e:
            raise ParameterPluginException(
                f"Parameter {action.name}/{parameter_id} has invalid value: '{param_value}'."
            ) from e

        if not isinstance(val, cls.type()):

            raise ParameterPluginException(f"Parameter {action.name}/{parameter_id} has invalid type: '{type(val)}'.")

        return val

    @classmethod
    def _value_from_json(cls, value: str) -> Any:

        try:
            return json.loads(value)
        except json.JsonException as e:
            raise ParameterPluginException(f"Invalid value '{value}'.") from e

    @classmethod
    def parameter_execution_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Any:

        # return copy in order to avoid unwanted changes in the original value if an action modifies the parameter
        return copy.deepcopy(cls.parameter_value(type_defs, scene, project, action_id, parameter_id))

    @classmethod
    @abc.abstractmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> AST:
        """Used for generating the main script logic.

        :param type_defs:
        :param scene:
        :param project:
        :param action_id:
        :param parameter_id:
        :return:
        """
        raise Arcor2NotImplemented()

    @classmethod
    def need_to_be_imported(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> None | list[ImportTuple]:
        return None

    @classmethod
    def value_to_json(cls, value: Any) -> str:
        return json.dumps(value)

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:
        return False

    @classmethod
    def uses_robot_joints(cls, project: CProject, action_id: str, parameter_id: str, robot_joints_id: str) -> bool:
        return False



import os

from arcor2.exceptions import Arcor2Exception


class Arcor2EnvException(Arcor2Exception):
    pass


def get_bool(variable_name: str, default: bool = False) -> bool:
    return os.getenv(variable_name, str(default)).lower() in ("true", "1")


def get_int(variable_name: str, default: None | int = None) -> int:

    val = os.getenv(variable_name, default)

    if val is None:
        raise Arcor2EnvException(f"Variable {variable_name} is not set.")

    try:
        return int(val)
    except ValueError:
        raise Arcor2EnvException(f"Variable {variable_name} has invalid value: {val}.")


def get_float(variable_name: str, default: None | float = None) -> float:

    val = os.getenv(variable_name, default)

    if val is None:
        raise Arcor2EnvException(f"Variable {variable_name} is not set.")

    try:
        return float(val)
    except ValueError:
        raise Arcor2EnvException(f"Variable {variable_name} has invalid value: {val}.")



import copy
from ast import Attribute, Load, Name
from typing import Any

from arcor2 import transformations as tr
from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import Position
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins import ParameterPluginException
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict


class PositionPlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return Position

    @classmethod
    def ap_id(cls, project: CProject, action_id: str, parameter_id: str) -> str:
        return cls._id_from_value(project.action(action_id).parameter(parameter_id).value)

    @classmethod
    def parameter_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Position:

        try:
            ap = project.bare_action_point(cls.ap_id(project, action_id, parameter_id))
        except Arcor2Exception as e:
            raise ParameterPluginException("Failed to get scene/project data.") from e
        return ap.position

    @classmethod
    def parameter_execution_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Position:

        # return copy in order to avoid unwanted changes in the original value if an action modifies the parameter
        return copy.deepcopy(tr.abs_position_from_ap(scene, project, cls.ap_id(project, action_id, parameter_id)))

    @classmethod
    def value_to_json(cls, value: Position) -> str:
        return value.to_json()

    @classmethod
    def parameter_ast(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Attribute:

        ap = project.bare_action_point(cls.ap_id(project, action_id, parameter_id))

        return Attribute(
            value=Attribute(
                value=Name(id="aps", ctx=Load()), attr=ap.name, ctx=Load()  # TODO this should not be hardcoded
            ),
            attr="position",  # TODO this should not be hardcoded
            ctx=Load(),
        )



#!/usr/bin/env python3

from arcor2.data.object_type import Box as BoxModel
from arcor2.object_types.upload import upload_def
from arcor2.test_objects.box import Box
from arcor2.test_objects.box2 import Box2
from arcor2.test_objects.dummy_multiarm_robot import DummyMultiArmRobot
from arcor2.test_objects.param_to_return import ParamToReturn
from arcor2.test_objects.position_param import PositionParam
from arcor2.test_objects.tester import Tester


def main() -> None:

    upload_def(PositionParam)
    upload_def(Box, BoxModel("Box", 0.1, 0.1, 0.1))
    upload_def(Box2, BoxModel("Box2", 0.2, 0.2, 0.2))
    upload_def(Tester, BoxModel("Tester", 0.3, 0.3, 0.3))
    upload_def(ParamToReturn)
    upload_def(DummyMultiArmRobot)


if __name__ == "__main__":
    main()




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/arcor2:arcor2_dist

from setuptools import setup

setup(**{
    'author': 'Robo@FIT',
    'author_email': 'imaterna@fit.vut.cz',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.10',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    'description': 'ARCOR2 base library',
    'entry_points': {
        'console_scripts': [
            'arcor2_upload_builtin_objects = arcor2.scripts.upload_builtin_objects:main',
        ],
    },
    'install_requires': (
        'Flask~=2.2.2',
        'Pillow~=9.4.0',
        'aiologger~=0.7.0',
        'apispec-webframeworks~=0.5.2',
        'apispec[yaml]~=6.0.2',
        'autopep8~=2.0.1',
        'colorlog~=6.7.0',
        'dataclasses-jsonschema[apispec,fast-dateparsing,fast-uuid,fast-validation]~=2.16.0',
        'fastuuid~=0.8.0',
        'flask-swagger-ui~=4.11.1',
        'flask_cors~=3.0.10',
        'numpy-quaternion[numba,scipy]~=2022.4.2',
        'numpy~=1.23.5',
        'opencv-contrib-python~=4.7.0.68',
        'orjson~=3.8.5',
        'packaging~=21.3',
        'pyhumps==3.8.0',
        'python-dateutil~=2.8.2',
        'requests~=2.28.2',
        'typing-inspect~=0.8.0',
        'urdfpy~=0.0.22',
        'websockets~=10.4',
    ),
    'license': 'LGPL',
    'long_description': """# arcor2
# Changelog

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),

## [1.0.0] - 2023-02-14

### Changed

- Updated dependencies.
- Compatibility with the Project service 1.0.0.
  - `Mesh` dataclass: `data_id` changed to `asset_id`.
  - Updated client (some paths were changed).

### Added

- Method `start` from the Scene service client now allows to specify a timeout. 

## [0.26.0] - 2022-12-12

### Changed

- Optional property `end_effector_id` added to `ProjectRobotJoints`.

## [0.25.0] - 2022-10-28

### Changed

- Switched to Python 3.10, updated dependencies.
- Removed unused `children` property from `SceneObject`.
- Removed unused assets-related functions from Project service client.

## [0.24.2] - 2022-08-29

### Changed

- REST client updated, so it is possible to specify a filename when, e.g., uploading a file to the Project service.

## [0.24.1] - 2022-08-17

### Changed

- Target REST endpoints for `arcor2.clients.scene_service` updated to match new API convention.
  - `/collisions/{model_type}/{model_id}` changed to `/collisions/{model_type}`, `model_id` moved to request 
  parameters as `{model_type}Id`.
  - `/transforms/{transform_id}/local-pose` changed to `/transforms/{transform_id}/pose/local`.
  - `/transforms/{transform_id}/world-pose` changed to `/transforms/{transform_id}/pose/world`.


## [0.24.0] - 2022-07-08

### Changed

- **BREAKING**: `data.common.WebApiError` is now JSON based exception.
  - Property `errors` was removed.
  - New properties `type`, `description` and optional `content`.

- **BREAKING**: `flask.FlaskException` class updated. 
  - `FlaskException` does no longer accept message and error code, but only `message` and optional `context`.
  - New `service` and `description` properties should be overridden by subclasses.
  - Exception instances are now convertible to JSON.

- **BREAKING**: Flask API error handlers return 500 error code for any error.
   `WebApiError` JSON object is supplied in response body.

- **BREAKING** : `rest.WebApiException`, `rest.RestHttpException` class gone. `WebApiError` and 
 `RestException` are raised instead. No HTTP code is supplied as the only possible error code is 500.

- **BREAKING** : `scene_service` client changed.
  - `world_pose` now call `.../world-pose` instead of `.../worldPose`.
  - `local_pose` now call `.../local-pose` instead of `.../localPose`.
  - `line_check` now call `.../line-safe` instead of `.../lineSafe`.
  - `upsert_collision` model id moved from query parameters to path.


### Added
- New `WebApiErrorFactory` class was implemented to generate service specific
  `WebApiError` schemas for swagger documentation.

- `flask.run_app` now accept optional `dependencies` argument.

## [0.23.1] - 2022-02-17

### Added

- Simple collision checking.
  - Related dataclasses were added.
  - New function in `scene_service` client.

## [0.23.0] - 2022-01-25

### Changed

- **BREAKING**: `ActionStateBefore` event updated.
  - Properties `action_id` and `parameters` are now optional.
  - New optional property `action_point_ids`. 
- Switched to Python 3.9, updated dependencies. 

## [0.22.0] - 2021-10-25

### Changed

- New abstract base class `VirtualCollisionObject`.
  - Meant as a base for 'dumb' collision objects, with primitive models, without any functionality, e.g. for safety 'walls' around the cell.
- `ActionMetadata` updated.
  - Removed useless properties (`blocking`, `blackbox`).
  - Added a new property `hidden` (UIs shall silently ignore those actions).
- Dropped support for `RelativePose` action parameter type.
- Handle all errors when importing stuff (`import_type_def`).
  - This is needed for instance in case of obsolete ObjectTypes.
  - ...when e.g. parameters of `ActionMetadata` were changed.
  - Otherwise outdated OT might raise `TypeError` or similar exceptions.
- New (optional) properties added to `BareActionPoint` (compatibility with Project service 0.14.0 or newer).
- Added `FlowActions` ObjectType, providing `is_first_pass` action which is useful to perform the initialization step.

### Fixed

- Action point transformations (reparenting).
  - When AP1 from `object -> AP1 -> AP2` was made global, AP2 was not updated properly.
  - When AP1 was made relative again, AP2 was again not updated.

## [0.21.0] - 2021-09-07

### Changed

- Added `Robot` specific exception `KinematicsException`.
  - So far used to give more precise error messages.
- `ProjectParameter` now derives from `Parameter`.
- `CachedScene` now caches ObjectTypes used in a scene.
- Logs are now colored.
- `Project` now has optional `project_objects_ids` property (not used at the moment).

### Fixed

- Make sure that `Pose` contain only floats.
  - When any value is e.g. np.float64, `orjson` complains.
  - Because of this, `to_dict` methods of `Position` and `Orientation` are overridden.
- Typo in URL within `get_models` of `project_service` client.
- Missing sleep in `wait_for` of `scene_service` client.

## [0.20.0] - 2021-08-05

### Changed

- Object hierarchy modified.
  - `GenericWithPose` now can't have a collision model.
  - A new `CollisionObject` must have collision model.
- `Robot` base class API changed.
  - Parameter `linear` was added to `move_to_pose`.
    - A robot not supporting linear movements should raise exception when `linear==True`.  
  - Parameter `include_gripper` was added to `robot_joints`.
    - By default, the method should return only arm's joints.
    - With the parameter set, the list should contain also gripper joints.
    - This is needed for visualization of grippers.
    - Makes only sense (so far) when gripper is part of URDF.
- Scene/Project clients now log underlying error messages.

### Fixed

- `rest` module now encodes body data as utf-8. 


## [0.19.0] - 2021-07-29

### Changed

- Support for multi arm robots.
  - New base class `MultiArmRobot`.
  - Its methods have an additional `arm_id` parameter.
  - There is a method to get arm IDs.
  - Relevant ARServer RPCs were extended with `arm_id`.
  - `ProjectRobotJoints` model now also contains `arm_id`.
  - There is `DummyMultiArmRobot` ObjectType for testing purposes.
- `CachedProject`: `constants` renamed to `parameters`.
- `Generic`: `INIT_PRIORITY` removed as useless.
- `GenericWithPose`: do not delete collision models in `cleanup`.
  - Collision models are removed by the Scene service on `stop`.
- Usage of `orjson` in `arcor2/json` and for dataclasses.
  - 40% speedup for serialization.
  - 80% speedup for deserialization.
  - Tested on `Project` dataclass.
- Compatibility with Project service 0.10.0.
- Compatibility with Scene service 0.5.0.  
- Uploading of meshes associated to ObjectTypes.
  - Parameter `file_to_upload` was added to `upload_def`.

### Fixed

- Correct default port for Project service (10000).

## [0.18.0] - 2021-06-14

### Changed
- Modules `package` and `resources` moved to `arcor2_execution_data`.

## [0.17.0] - 2021-06-11

### Changed
- Project service client updated to API version 0.8.0.
- `ARCOR2_PERSISTENT_STORAGE_URL` renamed to `ARCOR2_PROJECT_SERVICE_URL`.  
- Dependencies updated.

## [0.16.0] - 2021-05-21

### Changed
- Objects initialization order (`Resources`):
  - Object initialization order can be set using class-level `INIT_PRIORITY` variable.
  - The higher priority, the sooner are objects of that type initialized.
  - Objects are initialized serially.
- `CachedProject` has new methods to deal with hierarchy (`get_by_id`, `get_parent_id`, `childs`).
- Project service client updated to be compatible with version 0.7.0.
- Added a custom `json` module.
  - Wraps the standard json module, so it might be easier to replace it in the future.
  - Provides type annotations.
  - Raises a custom exception based on `Arcor2Exception`.

### Fixed
- Handling of context manager arguments in `Resources`.

## [0.15.0] - 2021-04-20

### Changed
- REST client now handles all codes >= 400 as errors.

## [0.14.1] - 2021-04-19

### Fixed
- Save and import of ObjectType was randomly failing because of [race condition](https://docs.python.org/3.8/library/importlib.html#importlib.machinery.FileFinder).

## [0.14.0] - 2021-03-30

### Changed
- Improved code for transforming poses.
- Use monkey patching instead of custom released version of dataclasses_jsonschema.
- Switched to builtin ast module instead of horast - much faster builds, etc.
- Added states `Stopping` and `Pausing` to `PackageState.Data.StateEnum`.
  - This is because it may take some time to pause or stop a package.
  - Other operations (as e.g. resume) are almost instant.

### Fixed
- `Resources` are now not sending `KeyboardInterrupt` as event.
- `import_type_def` now provides better error messages when import fails.
- Save and import of ObjectType was randomly failing.

## [0.13.0] - 2021-03-15

### Changed
- IDs (uuid) are now generated within the respective classes.
  - If necessary, ID can be still provided from outside as before.
  - IDs are prefixed so developers can easily check type of an object from its ID.
  - Prefix always start with a character.

### Fixed
- Composite actions are now properly handled by the `@action` decorator.

## [0.12.1] - 2021-03-08

### Fixed
- `Pose` parameter plugin fixed to generate correct code in a case when action on action point A uses orientation from action point B.

## [0.12.0] - 2021-03-03

### Fixed
- Parameter plugins now return copy of the parameter in order to prevent changes in the project if the value is modified e.g. within an action.
- Method `update_project_sources` of the Project service client was fixed.

### Changed
- Flask-based apps now don't log each API call by default.
  - It can be turned on by setting `ARCOR2_REST_API_DEBUG`.
- The `rest` module has a new exception type RestHttpException for getting HTTP error codes.
- `is_valid_identifier` now behaves the same as `is_valid_type`, it does not insist on convention (PascalCase vs snake_case) and provides concrete error messages.
- `Robot` API now has `safe` parameter.
- `Robot` now has API for hand teaching mode.  
- Line length of generated code changed from 80 to 120.

## [0.11.1] - 2021-02-09

### Fixed
- `@action` decorator fixed.
- `KeyError` was raised when `an` parameter was not given to an action.
- This only happened in the "manual" script writing scenario and when `patch_object_actions` was used.

## [0.11.0] - 2021-02-08

### Changed
- Explicit action parameters.
  - `Resources` class now do not need to deal with parameters.
  - Update of ObjectTypes (all actions now have the mandatory `an` parameter).
- `CurrentAction` removed, `ActionState` divided into `ActionStateBefore` and `ActionStateAfter`.
- New module with shared code for Flask-based apps.
- Updates of 3rd party dependencies.
- WS server now logs too long RPCs. 
  - Max. duration could be configured using `ARCOR2_MAX_RPC_DURATION`.

### Fixed
- At a startup, the main script now checks if the scene is running before attempt to stop it.
- `image_from_str` function fixed.

## [0.10.0] - 2020-12-14

### Changed
- `action` decorator now handles outputs of actions.
- `CachedProject` has new methods for handling project logic.
- `ActionState` event now contains action results.
- New built-in ObjectTypes `RandomActions`.
- Parameter plugins slightly reworked, some new helper functions.
- If a robot has URDF, it is zipped and uploaded to the Project service).
  - ...as a mesh file, which is workaround for missing storage of URDF models.
- Scene service client updated to support version 0.4.0.
- New built-in abstract ObjectType `Camera`.
- Initial support for loading of URDF files.
- Some first tests for parameter plugins were added.

## [0.9.2] - 2020-10-30

### Fixed
- `parse` function now also catches `ValueError` exception.
-  `check_object_type` now tries to parse whole module instead of just source of the class itself.

### Changed
- `Robot` base class now have methods for IK/FK.
- It is possible to use `==` on `Position` instances.

## [0.9.1] - 2020-10-19

### Fixed
- package_version was ignoring `package` argument

## [0.9.0] - 2020-10-16

### Changed
- Scene client updated for Scene 0.3.0.
- Exceptions refactored (BREAKING).
  - Arcor2Exception no longer has `message` property
  - There is a new package arcor2/exceptions.
  - Clients using `handle` decorator now have specific error messages defined.
  - Previously generated execution packages will become broken as `print_exception` function was moved 
- Loggers are now created using functions from `logging` module.
- `rest` module rewritten (BREAKING).
  - Now there is only one method `call` which takes http method as an argument.
  - There is `ARCOR2_REST_DEBUG` environment variable. When set, debugging logs are turned on.

### Fixed
- JSON containing only boolean value was not handled properly by `arcor2.rest` module.


## [0.8.0] - 2020-09-24
### Changed
- Reorganisation of the repository - switched to monorepo based on [Pants](https://www.pantsbuild.org/docs/welcome-to-pants). The code was divided into more packages (that can be separatelly relased) within one repository.
- Tests now run on GitHub instead of CircleCi.
- Unification of objects and services
  - There is ```Generic``` base class for objects without pose, ```GenericWithPose``` for objects with pose and ```Robot``` class that should be base class for every robot.
- Integration of scene service (0.2.0).
- @action decorator is now added automatically in the run-time.
- ```Orientation``` dataclass now performs quaternion normalization in ```__post_init__```.
- ```Robot``` base class now has ```_move_lock``` mutex to ensure that only one move-action is called at the time.

## [0.8.0rc2] - 2020-09-16

## [0.8.0rc1] - 2020-09-15

## [0.8.0b8] - 2020-08-21
### Fixed
- Some robot-related issues fixed

## [0.8.0b7] - 2020-08-12
### Changed
- Scene service client: 'upsert_collision' now has optional 'mesh_parameters': parameter.

## [0.8.0b6] - 2020-08-03
### Changed
- New logic representation
- Unification of objects and services
- Integration of scene service

## [0.7.1] - 2020-07-15
### Fixed
- Fix of broken python package arcor2 0.7.0

## [0.7.0] - 2020-07-15
### Changed

- ARServer: new RPC 'TemporaryPackage'
- ARServer: RPC ObjectTypesChangedEvent renamed to ChangedObjectTypesEvent, now contains changed ObjectTypes meta instead of just type names
- ARServer: ShowMainScreenEvent.
- Package name added to PackageInfoEvent
- ARServer now compares its API_VERSION with API_VERSION of the Execution.
- ARServer: ShowMainScreenEvent will now not contain 'highlight' when sent to a newly connected UI.
- AP can now have another AP as parent
- rest: OptionalData now may contain list of primitive type.
- Execution: PackageStateEvent now contains package_id
- Execution: added 'executed' to PackageMeta

## [0.6.0] - 2020-06-19
### Changed

- Build/Execution proxy: allow port change using env. var.
- ARServer: RenameScene RPC now checks if scene name is unique and 'dry_run' works.
- ARServer: ListScenes/ListProjects now contain 'modified'.
- ARServer: DeleteObjectType RPC added.
- @action decorator is now compatible with Windows.
- Service class now has 'cleanup' method which is called when scene is closed or when script ends. Call of 'cleanup' can be disabled by 'export ARCOR_CLEANUP_SERVICES=False' - this is particularly useful when running the script manually again and again.
- Cleanup method for ObjectTypes.

## [0.5.1] - 2020-06-04
### Fixed
- ignoring check of return parameters
- allowing list of strings as request body

## [0.5.0] - 2020-06-01
### Changed
- ARServer container need to setup new env variable using docker-compose -> ARCOR2_DATA_PATH=/root/data
- ListProjects RPC now gets projects in parallel.
- dry_run parameter for selected RPCs
- EEF pose/robot joints streaming
- OpenScene, OpenProject, SceneClosed, ProjectClosed events.
- Execution proxy: use persistent websocket connection.
- SceneCollisionsEvent merged into PackageInfoEvent
- ARServer: RPC to cancel action execution.
- Execution package now contains package.json with its metadata. Execution service now supports renaming of packages.

## [0.4.3] - 2020-05-22
### Changed
- added support for CORS

## [0.4.2] - 2020-04-27
### Fixed
- Fix of functions to transform relative poses to absolute and vice versa

## [0.4.1] - 2020-04-22
### Added
- New RPCs for getting robot joints and effector pose
- New RPC to get IDs of EE and suctions
- Added pivot enum for UpdateObjectPoseUsingRobot

### Fixed
- Fix of remove action RPC
- Another fixes


## [0.4.0] - 2020-04-17
### Changed
- Complete redesign of RPC for ARClients (AREditor atm)
- Documentation of execution and build API
- Support for project service 0.2.0
- New and updated events
- Enhanced error messages
- Create (global) session to enable reuse of connections.

## [0.3.0] - 2020-03-24
### Changed
- Renamed RobotJoints to ProjectRobotJoints and ModelTypeEnum to Model3dType
- Added new services for Time and Logic related actons
- Added boolean parameter plugin
- Description, returns and origins fields marked as optional
- New event - ActionResult
- Separated script enabling discovery through UDP broadcast
- Support for list params
- Services and Action objects are now marked as disabled when some problem occured and error message is passed to GUI (previously such services/objects were ignored)
- Services with no configuration are disabled

## [0.2.1] - 2020-02-28
### Fixed
- Added compatibility with Project service v 0.1.1
- Param values and defaults are strings now
- min, max stored in extra as JSON string

## [0.2.0] - 2020-02-20
### Changed
- ExecuteAction RPC.
- Uuid for action object/point/action.
- Execution proxy PUT method
- ActionPoint class in execution package
- Removed loop in main script, when hasLogic == false
- Parameter values not send in currentAction event
- ProjectState RESUMED removed
- Execution: print out script output if not JSON.
- Joint: rotation -> value


## [0.1.7] - 2019-12-24
### Fixed
- Build: disable caching

## [0.1.5] - 2019-12-22
### Fixed
- Parameter plugins

## [0.1.4] - 2019-12-18
### Fixed
- Parameter of type relative_pose now accepts json string as value

## [0.1.3] - 2019-12-18
### Fixed
- N/A

## [0.1.2] - 2019-12-17
### Fixed
- bump docker version

## [0.1.1] - 2019-12-17
### Fixed
- bump docker version

## [0.1.1] - 2019-12-12
### Fixed
- N/A

## [0.1.0] - 2019-12-12
### Changed
- Separation of services.""",
    'long_description_content_type': 'text/markdown',
    'name': 'arcor2',
    'namespace_packages': (
    ),
    'package_data': {
        'arcor2': (
            'VERSION',
            'py.typed',
        ),
    },
    'packages': (
        'arcor2',
        'arcor2.clients',
        'arcor2.data',
        'arcor2.data.rpc',
        'arcor2.exceptions',
        'arcor2.object_types',
        'arcor2.parameter_plugins',
        'arcor2.scripts',
        'arcor2.source',
        'arcor2.test_objects',
    ),
    'python_requires': '==3.10.*',
    'version': '1.0.0',
})



from dataclasses import dataclass

import numpy as np
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.exceptions import Arcor2Exception


@dataclass
class CameraParameters(JsonSchemaMixin):

    fx: float
    fy: float
    cx: float
    cy: float
    dist_coefs: list[float]  # (k1,k2,p1,p2[,k3[,k4,k5,k6],[s1,s2,s3,s4]])

    def __post_init__(self) -> None:

        if len(self.dist_coefs) not in (4, 5, 8, 12):
            raise Arcor2Exception("Unsupported number of distortion coefficients.")

    def as_camera_matrix(self) -> np.ndarray:

        return np.array(
            [
                [self.fx, 0.00000, self.cx],
                [0.00000, self.fy, self.cy],
                [0.00000, 0.00000, 1],
            ]
        )
